from argparse import Namespace
from collections import Counter
from enum import auto
from fnmatch import fnmatch
from pathlib import Path
from typing import Dict, List, Tuple, Union

import typer
from strenum import StrEnum
from tabulate import tabulate

from nef_pipelines.lib.constants import NEF_UNKNOWN
from nef_pipelines.lib.nef_lib import (
    loop_row_dict_iter,
    loop_row_namespace_iter,
    read_entry_from_stdin_or_exit,
)
from nef_pipelines.lib.structures import AtomLabel, Peak, PeakValues, SequenceResidue
from nef_pipelines.lib.util import exit_error, is_float, parse_comma_separated_options
from nef_pipelines.transcoders.nmrview import export_app

app = typer.Typer()


class HEADINGS(StrEnum):
    LABEL = auto()
    SHIFT = auto()
    WIDTH = auto()
    BOUND = auto()
    ERROR = auto()
    COUPLING = auto()
    COMMENT = auto()


HEADING_TRANSLATIONS = {
    HEADINGS.LABEL: "L",  # the assignment
    HEADINGS.SHIFT: "P",  # the shift
    HEADINGS.WIDTH: "W",  # width at half height in ppm
    HEADINGS.BOUND: "B",  # bound width at lowest contour level when picked in ppm, assume gaussian?
    HEADINGS.ERROR: "E",  # an error code always '++' ?
    HEADINGS.COUPLING: "J",  # always 0.00 for us
    HEADINGS.COMMENT: "U",  # user comment always currently {?}
}


# noinspection PyUnusedLocal
@export_app.command()
def peaks(
    file_name_template: str = typer.Option(
        "%s",
        help="the template for the filename to export to %s will get replaced by the axis_name of the peak frame",
        metavar="<peak-file.xpk>",
    ),
    frame_selectors: List[str] = typer.Argument(
        None, help="the names of the frames to export", metavar="<frames>"
    ),
):
    frame_selectors = parse_comma_separated_options(frame_selectors)

    if len(frame_selectors) == 0:
        frame_selectors = [
            "*",
        ]

    entry = read_entry_from_stdin_or_exit()

    SPECTRUM_CATEGORY = "nef_nmr_spectrum"
    peaks = entry.get_saveframes_by_category(SPECTRUM_CATEGORY)

    names_and_frames = {
        frame.name[len(SPECTRUM_CATEGORY) :].lstrip("_"): frame for frame in peaks
    }

    selected_frames = {}
    for frame_selector in frame_selectors:
        selection = {
            name: frame
            for name, frame in names_and_frames.items()
            if fnmatch(name, frame_selector)
        }

        selected_frames.update(selection)

    if len(selected_frames) == 0:
        for frame_selector in frame_selectors:
            selection = {
                name: frame
                for name, frame in names_and_frames.items()
                if fnmatch(name, f"*{frame_selector}*")
            }

            selected_frames.update(selection)

    if not "%s" and len(selected_frames) > 1:
        exit_error(
            f"%s is not in the filename template and there is more than one file, template {file_name_template}"
        )

    # bad_characters = '''\
    #     `:';~@$%^&().<>"
    # '''
    # replacements = '_' * len(bad_characters)
    # translation_table = str.maketrans(bad_characters, replacements)

    for frame_name, frame in selected_frames.items():
        # peak_list_name = f'''{frame_name.translate(translation_table).strip('_')}.xpk'''

        SPECTRUM_DIMESIONS = "_nef_spectrum_dimension"
        PEAKS = "_nef_peak"
        spectrum_dimensions = list(loop_row_namespace_iter(frame[SPECTRUM_DIMESIONS]))

        spectrum_name = (
            frame["ccpn_spectrum_file_path"][0]
            if "ccpn_spectrum_file_path" in frame
            else "unknown"
        )
        spectrum_name = Path(spectrum_name).parts[-1]

        result = ["label dataset sw sf"]

        axis_names = [
            dimension.axis_code
            if "axis_code" in dimension and dimension != NEF_UNKNOWN
            else "unknown"
            for dimension in spectrum_dimensions
        ]

        axis_names = _make_names_unique(axis_names)

        result.append(" ".join(axis_names))

        # TODO: issue warning if spectrum name doesn't end in nv!
        nmrview_spectrum_name = Path(spectrum_name)
        nmrview_spectrum_name = f"{nmrview_spectrum_name.stem}.nv"
        result.append(nmrview_spectrum_name)

        spectrometer_frequencies = [
            dimension.spectrometer_frequency for dimension in spectrum_dimensions
        ]
        spectrometer_frequencies = _convert_to_floats_or_exit(
            spectrometer_frequencies, "spectrometer_frequencies"
        )

        sweep_widths = [dimension.spectral_width for dimension in spectrum_dimensions]
        sweep_widths = _convert_to_floats_or_exit(sweep_widths, "sweep_widths")

        sweep_widths = [
            sweep_width * spectrometer_frequency
            for sweep_width, spectrometer_frequency in zip(
                sweep_widths, spectrometer_frequencies
            )
        ]

        sweep_widths = [f"{{{sweep_width:.3f}}}" for sweep_width in sweep_widths]
        result.append(" ".join(sweep_widths))

        spectrometer_frequencies = [
            f"{{{spectrometer_frequency:.3f}}}"
            for spectrometer_frequency in spectrometer_frequencies
        ]
        result.append(" ".join(spectrometer_frequencies))

        print("\n".join(result))

        headings = [
            f"{axis_name}.{item}" for axis_name in axis_names for item in "LPWBEJU"
        ]
        headings.extend("vol int stat comment flag0".split())
        headings.insert(0, "")

        table = [headings]

        pipeline_peaks = []
        for peak_row in list(loop_row_dict_iter(frame[PEAKS])):
            pipeline_peaks.append(_row_to_peak(axis_names, peak_row))

        for peak in pipeline_peaks:

            row = _build_nmrview_row(peak, axis_names)
            table.append(row)

        print(tabulate(table, tablefmt="plain"))


def _peak_to_atom_labels(row: Namespace) -> List[AtomLabel]:
    result = []
    for i in range(1, 15):
        dim_chain_code = f"chain_code_{i}"
        dim_sequence_code = f"sequence_code_{i}"
        dim_residue_name = f"residue_name_{i}"
        dim_atom_name = f"atom_name_{i}"

        if (
            dim_chain_code in row
            and dim_sequence_code in row
            and dim_residue_name in row
            and dim_atom_name in row
        ):
            chain_code = getattr(row, dim_chain_code)
            sequence_code = getattr(row, dim_sequence_code)
            residue_name = getattr(row, dim_residue_name)
            atom_name = getattr(row, dim_atom_name)

            result.append(
                AtomLabel(
                    SequenceResidue(chain_code, sequence_code, residue_name), atom_name
                )
            )

    return result


def _atom_label_to_nmrview(label: AtomLabel) -> str:
    return f"{{{label.residue.sequence_code}.{label.atom_name}}}"


def _convert_to_floats_or_exit(
    putative_floats: List[str], thing_types: str
) -> List[float]:
    are_floats = [
        is_float(spectrometer_frequency) for spectrometer_frequency in putative_floats
    ]
    if not all(are_floats):
        bad_dims = [
            dim
            for dim, is_float_value in enumerate(are_floats, start=1)
            if not is_float_value
        ]
        bad_values = [
            value
            for is_float_value, value in zip(are_floats, putative_floats)
            if not is_float_value
        ]
        message = f"""\
                the following {thing_types} frequencies which cannot be converted to floats
                dims: {', '.join(bad_dims)}
                bad values: {', '.join(bad_values)}
            """

        exit_error(message)

    return [float(value) for value in putative_floats]


def _row_to_peak(
    axis_names: Tuple[str], row: Dict[str, Union[int, str, float]]
) -> Peak:
    peak_id = row["peak_id"]
    volume = row["volume"] if row["volume"] != NEF_UNKNOWN else None
    volume_uncertainty = (
        row["volume_uncertainty"] if row["volume_uncertainty"] != NEF_UNKNOWN else None
    )
    height = row["height"] if row["height"] != NEF_UNKNOWN else None
    height_uncertainty = (
        row["height_uncertainty"] if row["height_uncertainty"] != NEF_UNKNOWN else None
    )

    positions = {}
    position_uncertainties = {}
    for axis_index, axis_name in enumerate(axis_names, start=1):
        position_tag = f"position_{axis_index}"
        position_uncertainty_tag = f"position_uncertainty_{axis_index}"
        positions[axis_name] = row[position_tag]
        position_uncertainties[axis_name] = (
            row[position_uncertainty_tag]
            if row[position_uncertainty_tag] != NEF_UNKNOWN
            else None
        )
        position_uncertainties[axis_name]

    assignments = {}
    for axis_index, axis_name in enumerate(axis_names, start=1):
        assignment_chain_code_tag = f"chain_code_{axis_index}"
        assignment_sequence_code_tag = f"sequence_code_{axis_index}"
        assignment_residue_name_tag = f"residue_name_{axis_index}"
        assignment_atom_name_tag = f"atom_name_{axis_index}"

        chain_code = row[assignment_chain_code_tag]
        sequence_code = row[assignment_sequence_code_tag]
        residue_name = row[assignment_residue_name_tag]
        atom_name = row[assignment_atom_name_tag]

        is_assigned = (
            chain_code != NEF_UNKNOWN
            and sequence_code != NEF_UNKNOWN
            and atom_name != NEF_UNKNOWN
        )
        atom_labels = (
            [
                AtomLabel(
                    SequenceResidue(chain_code, sequence_code, residue_name), atom_name
                ),
            ]
            if is_assigned
            else []
        )

        assignments[axis_name] = atom_labels

    uncertainties_none = [
        uncertainty is None for uncertainty in position_uncertainties.values()
    ]
    if all(uncertainties_none):
        position_uncertainties = None

    peak_values = PeakValues(
        serial=peak_id + 1,
        volume=volume,
        height=height,
        volume_uncertainty=volume_uncertainty,
        height_uncertainty=height_uncertainty,
    )
    return Peak(
        id=peak_id,
        values=peak_values,
        positions=positions,
        position_uncertainties=position_uncertainties,
        assignments=assignments,
    )


def _peak_to_nmrview_label(atom_label: AtomLabel, default_chain=None):
    result = []

    nmrview_chain = (
        atom_label.residue.chain_code
        if atom_label.residue.chain_code != NEF_UNKNOWN
        else default_chain
    )
    if nmrview_chain is not None:
        result.append(nmrview_chain)

    nmrview_residue = (
        atom_label.residue.sequence_code
        if atom_label.residue.sequence_code != NEF_UNKNOWN
        else "{?}"
    )

    result.append(str(nmrview_residue))

    nmrview_atom = atom_label.atom_name
    result.append(nmrview_atom)

    return f'{{{".".join(result)}}}'


def _build_nmrview_row(peak, axis_names):

    # TODO: are we handling peak serials correctly...
    result = [peak.id]
    nmrview_assignments = _peak_to_assignments(peak, axis_names)
    for axis_name in axis_names:
        for heading, letter in HEADING_TRANSLATIONS.items():
            if heading == HEADINGS.LABEL:
                assignment = nmrview_assignments[axis_name]
                if len(assignment.strip()) == 0:
                    assignment = "{?}"
                result.append(assignment)

            if heading == HEADINGS.SHIFT:
                result.append(f"{peak.positions[axis_name]:.3f}")

            # TODO need default values per isotope and some calculations
            if heading == HEADINGS.WIDTH:
                # TODO: place holder
                result.append("0.024")

            # TODO need default values per isotope and some calculations or
            if heading == HEADINGS.BOUND:
                # TODO: place holder
                result.append("0.051")

            if heading == HEADINGS.ERROR:
                # TODO: is this the correct value for a correct peak
                # most probably the best we can do is presume all peaks are good
                result.append("++")

            if heading == HEADINGS.COUPLING:
                result.append("0.000")

            if heading == HEADINGS.COMMENT:
                # TODO: is there a peak axis comment in ccpn/nef
                result.append("{?}")

    volume = peak.values.volume if peak.values.volume else 0.000
    result.append(f"{volume:.3f}")

    height = peak.values.height if peak.values.height else 0.000
    result.append(f"{height:.3f}")

    # status
    result.append("0")

    if peak.values.comment != "":
        result.append(f"{{{peak.values.comment}}}")
    else:
        result.append("{?}")

    # flag0
    result.append("0")

    return result


def _peak_to_assignments(peak: Peak, axis_names: Tuple[str]) -> List[str]:
    # print(peak)
    num_assignments = [len(peak.assignments[axis]) for axis in axis_names]
    max_num_assignments = max(num_assignments)

    # print('assignments', peak.assignments)

    if max_num_assignments == 0:
        result = {axis_name: "{?}" for axis_name in axis_names}
    else:
        result = {}
        for axis_name in axis_names:
            result[axis_name] = " ".join(
                [
                    _peak_to_nmrview_label(atom_label)
                    for atom_label in peak.assignments[axis_name]
                ]
            )

    return result


def _make_names_unique(axis_names: List[str]) -> List[str]:

    seen = Counter()
    for axis_name in axis_names:
        seen[axis_name] += 1

    result = []
    modified = Counter()
    for axis_name in axis_names:
        if seen[axis_name] > 1:
            modified[axis_name] += 1
            axis_name = f"{axis_name}_{modified[axis_name]}"
        result.append(axis_name)

    return result



import pytest

from nef_pipelines.transcoders.nmrview import nmrview_lib


@pytest.fixture
def parser():

    return nmrview_lib.get_tcl_parser()


@pytest.fixture
def parse_tcl():

    return nmrview_lib.parse_tcl


def test_basic_word(parser):
    values = ["123", "abc", "123.abc", "123.abc#"]

    for value in values:
        parsed = parser.parseString(value)

        assert len(parsed) == 1
        assert parsed[0] == value


def test_quoted_word(parser):
    values = ['"123"', '"abc"', '"123.abc"']

    for value in values:
        parsed = parser.parseString(value)

        assert len(parsed) == 1
        assert parsed[0] == value.strip('"')


def test_quoted_word_list(parser):
    values = ['"123 456"', '"abc def"', '"123.abc 456.def"']

    for value in values:
        parsed = parser.parseString(value)

        assert len(parsed) == 2
        assert list(parsed) == value.strip('"').split()


def test_string_and_quoted_word_list(parser):
    tests_and_expecteds = {
        '"abc" "123 456"': ["abc", ["123", "456"]],
        '"" "123 456"': ["", ["123", "456"]],
    }

    for test, expected in tests_and_expecteds.items():
        parsed = parser.parseString(test)

        assert list(parsed.asList()) == expected


def test_complex_list(parser):
    tests_and_expecteds = {
        "{123 456}": ["123", "456"],
        '"" {123 456}': ["", ["123", "456"]],
        "{} {123 456}": ["", ["123", "456"]],
        "{} {{} 123 456}": ["", [[], "123", "456"]],
        "{} {{} 123 {} 456}": ["", [[], "123", [], "456"]],
        '{} "{} 123 {} 456"': ["", [[], "123", [], "456"]],
    }

    for test, expected in tests_and_expecteds.items():
        parsed = parser.parseString(test)

        assert list(parsed.asList()) == expected


def test_rest_of_line(parse_tcl, capsys):
    test = "12{3"

    with pytest.raises(SystemExit):
        parse_tcl(test, file_name="Wibble.txt", line_no=666)

    stderr = capsys.readouterr().err
    print(stderr)
    assert "Expected end of text, found '{'" in stderr
    assert "line: 667" in stderr
    assert "file: Wibble.txt" in stderr


if __name__ == "__main__":
    pytest.main([__file__, "-vv"])



#!/usr/bin/env python3
# for pynmrstar to avoid WARNING:root:Loop with no data on line: xx
import logging
import sys
import traceback
from textwrap import dedent

try:
    import future  # noqa: F401
except Exception:
    pass

MINIMUM_VERSION = (3, 7)
if not sys.version_info >= MINIMUM_VERSION:
    print(
        "Error; minimum version required id %i.%i [and python 2 is not supported]"
        % MINIMUM_VERSION,
        file=sys.stderr,
    )
    print("exiting...", file=sys.stderr)


logging.getLogger().setLevel(logging.ERROR)

EXIT_ERROR = 1


def do_exit_error(msg, trace_back=True, exit_code=EXIT_ERROR):
    msg = dedent(msg)
    if trace_back:
        traceback.print_exc()
    print(msg, file=sys.stderr)
    print("exiting...", file=sys.stderr)
    sys.exit(exit_code)


def create_nef_app():
    import typer

    from nef_pipelines import nef_app

    nef_app.app = typer.Typer(no_args_is_help=True)
    app = nef_app.app  # noqa: F841
    return nef_app


def main():
    try:
        import typer
        from click import ClickException

    except Exception as e:

        msg = """\

             Initializaion error: one of the core libraries [friendly_tracback/typer] if missing from you environment
             please make sure they are installed an try again
             exiting..."""
        do_exit_error(msg, e)

    try:
        nef_app = create_nef_app()

    except Exception as e:
        msg = """\

                 Initialisation error: failed to start the typer app, message the developer
                 exiting..."""
        do_exit_error(msg, e)

    try:
        # import components which will self register, this could and will be automated
        import nef_pipelines.tools.about  # noqa: F401
        import nef_pipelines.tools.chains  # noqa: F401
        import nef_pipelines.tools.entry  # noqa: F401
        import nef_pipelines.tools.frames  # noqa: F401
        import nef_pipelines.tools.header  # noqa: F401
        import nef_pipelines.tools.peaks  # noqa: F401
        import nef_pipelines.tools.shifts  # noqa: F401
        import nef_pipelines.tools.stream  # noqa: F401
        import nef_pipelines.tools.test  # noqa: F401
        import nef_pipelines.transcoders.csv  # noqa: F401
        import nef_pipelines.transcoders.fasta  # noqa: F401
        import nef_pipelines.transcoders.mars  # noqa: F401
        import nef_pipelines.transcoders.nmrpipe  # noqa: F401
        import nef_pipelines.transcoders.nmrview  # noqa: F401
        import nef_pipelines.transcoders.pales  # noqa: F401
        import nef_pipelines.transcoders.pdbx  # noqa: F401
        import nef_pipelines.transcoders.rpf  # noqa: F401
        import nef_pipelines.transcoders.shifty  # noqa: F401
        import nef_pipelines.transcoders.sparky  # noqa: F401
        import nef_pipelines.transcoders.xcamshift  # noqa: F401
        import nef_pipelines.transcoders.xplor  # noqa: F401

    except Exception as e:
        msg = """\

             Initialisation error: failed to load a plugin, remove the plugin or contact the developer
             """

        do_exit_error(msg, e)

    try:

        nef_app.app

        command = typer.main.get_command(nef_app.app)

        command(standalone_mode=False)

    except ClickException as e:
        e.show()
        do_exit_error(
            f"inputs: {' '.join(sys.argv[1:])}", trace_back=False, exit_code=e.exit_code
        )

    except Exception as e:

        msg = f"""\

              Runtime error: failed to process the data using the plugin and commands, check you inputs or report a bug
              inputs: {' '.join(sys.argv[1:])}
              message: {e}
              """

        do_exit_error(msg)


if __name__ == "__main__":
    main()



import typer
from typer.testing import CliRunner

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.tools.frames.tabulate import tabulate

runner = CliRunner()
app = typer.Typer()
app.command()(tabulate)


# noinspection PyUnusedLocal
def test_frame_basic(clear_cache):

    path = path_in_test_data(__file__, "tailin_seq_short.nef")

    result = run_and_report(app, ["--pipe", path])

    if result.exit_code != 0:
        print("INFO: stdout from failed read:\n", result.stdout)

    assert result.exit_code == 0

    EXPECTED = """\
      ind   chain       seq   resn     link
         1  A             10  GLU      start
         2  A             11  TYR      middle
         3  A             12  ALA      middle
         4  A             13  GLN      middle
         5  A             14  PRO      middle
         6  A             15  ARG      middle
         7  A             16  LEU      middle
         8  A             17  ARG      middle
         9  A             18  LEU      middle
        10  A             19  GLY      middle
        11  A             20  PHE      middle
        12  A             21  GLU      middle
        13  A             22  ASP      end

    """

    assert_lines_match(EXPECTED, result.stdout)



import typer

from nef_pipelines import nef_app

app = typer.Typer()
import_app = typer.Typer()


if nef_app.app:

    nef_app.app.add_typer(app, name="csv", help="- read [rdcs]")

    app.add_typer(import_app, name="import", help="- import [rdcs]")

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.transcoders.csv.importers.rdcs  # noqa: F401



from collections import UserList
from functools import partial
from pathlib import Path
from textwrap import dedent
from typing import Dict, List, Tuple

from pynmrstar import Loop, Saveframe
from pyparsing import (
    CaselessLiteral,
    Combine,
    Forward,
    Group,
    Literal,
    OneOrMore,
    Optional,
    Or,
    ParseException,
    ParseResults,
    Regex,
    Suppress,
    Word,
    ZeroOrMore,
    alphanums,
    nums,
)
from pyparsing.common import pyparsing_common as ppc

from nef_pipelines.lib.nef_lib import UNUSED, PotentialTypes
from nef_pipelines.lib.sequence_lib import ANY_CHAIN, replace_chain_in_atom_labels
from nef_pipelines.lib.structures import (
    AtomLabel,
    DihedralRestraint,
    DistanceRestraint,
    SequenceResidue,
)
from nef_pipelines.lib.util import (
    end_with_ordinal,
    exit_error,
    get_display_file_name,
    read_from_file_or_exit,
)

# trimmed [minimal set of factors supporting most restraints] EBNF for distance restraints from xplor manual
# <selection>:== ( <selection-expression> )
# <selection-expression>:==
#   <term> - selects atoms that belong to the term.
#   <term> { OR <term> } selects all atoms that belong to either one of the terms.
#
# <term>:==
#   <factor> - selects atoms that belong to the factor.
#   <factor> { AND <factor> } - selects all atoms that belong to all of the factors.
#
# <factor>:==
#   ( <selection-expression> ) - selects all atoms that are selected in selection expression.
#
#   ALL - selects all atoms.
#
#   ATOM <*segment-name*> <*residue-number*> <*atom*>
#
#   NAME <*atom*> - selects all atoms that match the specified atom name (Section 3.1.1)
#                       or a wildcard (Section 2.6) of it.
#
#   RESIdue <*residue-number*> - selects all atoms that match the specified residue number (Section 3.7)
#                                    or a wildcard (Section 2.6) of it.
#
#   SEGIdentifier <*segment-name*> selects all atoms that match the specified segment name (Section 3.7)
#                                  or a wildcard (Section 2.6) of it.

LEFT_PARENTHESIS = "("
RIGHT_PARENTHESIS = ")"

ASSIGN = "assign"
NEF_DIHEDRAL_RESTRAINT = "nef_dihedral_restraint"
XPLOR_COMMENT_TOKEN = "!"
DISTANCE_PLUS = "d_plus"
DISTANCE_MINUS = "d_minus"
DISTANCE = "d"
EXPONENT = "exponent"
RANGE = "range"
ANGLE = "angle"
ENERGY_CONSTANT = "energy_constant"

ATOMS_4 = "atoms_4"
ATOMS_3 = "atoms_3"
ATOMS_2 = "atoms_2"
ATOMS_1 = "atoms_1"

SELECTION = "selection"
SELECTION_EXPRESSION = "selection-expression"
TERM = "term"
FACTOR = "factor"

AND = "AND"
OR = "OR"

SEGMENT_FACTOR = "segment-factor"
RESIDUE_FACTOR = "residue-factor"
ATOM_FACTOR = "atom-factor"

RESIDUE_LITERAL = "residue"
SEGIMENT_IDENTIFIER_LITERAL = "segidentifier"
ATOM_NAME_LITERAL = "name"

ATOM_WILDCARDS = "*%#+"

ATOM = "atom"
SEGID = "segid"
RESID = "resid"

XPLOR_COMMENT = Regex(r"!.*").set_name("Xplor comment")

DEFAULT_PRECISION = 5
INDEX = "index"
RESTRAINT_ID = "restraint_id"
RESTRAINT_COMBINATION_ID = "restraint_combination_id"
CHAIN_CODE_1 = "chain_code_1"
SEQUENCE_CODE_1 = "sequence_code_1"
RESIDUE_NAME_1 = "residue_name_1"
ATOM_1 = "atom_name_1"
CHAIN_CODE_2 = "chain_code_2"
SEQUENCE_CODE_2 = "sequence_code_2"
RESIDUE_NAME_2 = "residue_name_2"
ATOM_2 = "atom_name_2"
CHAIN_CODE_3 = "chain_code_3"
SEQUENCE_CODE_3 = "sequence_code_3"
RESIDUE_NAME_3 = "residue_name_3"
ATOM_3 = "atom_name_3"
CHAIN_CODE_4 = "chain_code_4"
SEQUENCE_CODE_4 = "sequence_code_4"
RESIDUE_NAME_4 = "residue_name_4"
ATOM_4 = "atom_name_4"
WEIGHT = "weight"
TARGET_VALUE = "target_value"
LOWER_LIMIT = "lower_limit"
UPPER_LIMIT = "upper_limit"

DISTANCE_RESTRAINT_TAGS = [
    INDEX,
    RESTRAINT_ID,
    RESTRAINT_COMBINATION_ID,
    CHAIN_CODE_1,
    SEQUENCE_CODE_1,
    RESIDUE_NAME_1,
    ATOM_1,
    CHAIN_CODE_2,
    SEQUENCE_CODE_2,
    RESIDUE_NAME_2,
    ATOM_2,
    WEIGHT,
    TARGET_VALUE,
    LOWER_LIMIT,
    UPPER_LIMIT,
]

DIHEDRAL_RESTRAINT_TAGS = [
    INDEX,
    RESTRAINT_ID,
    RESTRAINT_COMBINATION_ID,
    CHAIN_CODE_1,
    SEQUENCE_CODE_1,
    RESIDUE_NAME_1,
    ATOM_1,
    CHAIN_CODE_2,
    SEQUENCE_CODE_2,
    RESIDUE_NAME_2,
    ATOM_2,
    CHAIN_CODE_3,
    SEQUENCE_CODE_3,
    RESIDUE_NAME_3,
    ATOM_3,
    CHAIN_CODE_4,
    SEQUENCE_CODE_4,
    RESIDUE_NAME_4,
    ATOM_4,
    WEIGHT,
    TARGET_VALUE,
    LOWER_LIMIT,
    UPPER_LIMIT,
]


class XPLORParseException(Exception):
    ...


_and_op = CaselessLiteral(AND)
_or_op = CaselessLiteral(OR)

_l_paren = Suppress(LEFT_PARENTHESIS)
_r_paren = Suppress(RIGHT_PARENTHESIS)


def _expand_literal(literal, min_length=4, case_insenitive=True):
    result = []
    for i in range(min_length, len(literal) + 1):
        elem = CaselessLiteral(literal[:i]) if case_insenitive else Literal(literal[:i])
        result.append(elem)

    result.reverse()

    return Or(result)


def _convert_to_4_letter_string(string: str) -> str:
    result = string
    if string_len := len(string) < 4:
        result += " " * (4 - string_len)

    return result


class _NamedToken:
    def __init__(self, name, toks):
        self.name = name
        self.token = toks

    def getName(self):
        return self.name

    get_name = getName

    def __len__(self):
        return 1

    def __getitem__(self, item):
        if item == 0:
            return self.token
        else:
            raise Exception("array out of bounds")

    def __str__(self):
        return f"{self.name} : {self.token}"

    def __repr__(self):
        return f"{self.name} : {self.token}"

    def __eq__(self, other):
        result = False
        if isinstance(other, self.__class__):
            if other.name == self.name and other.token == self.token:
                result = True

        return result


class _NamedTokens(UserList):
    def __init__(self, name, toks):

        super(_NamedTokens, self).__init__(toks)

        self.name = name

    def __str__(self):
        return f"[{self.name} : {self.data}]"

    def _repr__(self):
        tok_strings = ",".join([str(tok) for tok in self.toks])
        return f"{self.name} {tok_strings}"


def _as_named_single_token(name, s, loc, toks):
    return _NamedToken(name, toks[0])


# resid factor
_residue_literal = Suppress(_expand_literal(RESIDUE_LITERAL))
_residue_number = (
    Combine((Optional("-") + Word(nums)))
    .set_results_name(RESID, list_all_matches=True)
    .setParseAction(ppc.convert_to_integer)
)(RESID)
_residue_factor = (_residue_literal + _residue_number)(RESIDUE_FACTOR)

_named_resid = partial(_as_named_single_token, RESID)
_residue_factor.set_parse_action(_named_resid)

# atom factor
_atom_literal = Suppress(CaselessLiteral(ATOM_NAME_LITERAL))
_atom_label = Word(alphanums + ATOM_WILDCARDS).set_results_name(
    ATOM, list_all_matches=True
)
_atom_factor = (_atom_literal + _atom_label)(ATOM_FACTOR)

_named_atom = partial(_as_named_single_token, ATOM)
_atom_factor.set_parse_action(_named_atom)

# segment factor
_segid_literal = Suppress(_expand_literal(SEGIMENT_IDENTIFIER_LITERAL))
_segid_label = Word(alphanums, max=4)
_segid_factor = (_segid_literal + _segid_label)(SEGMENT_FACTOR)


_named_segid = partial(_as_named_single_token, SEGID)
_segid_factor.setParseAction(_named_segid)


def _as_named_tokens(name, s, loc, toks):
    return _NamedTokens(name, toks)


_as_named_list_or = partial(_as_named_tokens, OR)
_as_named_list_and = partial(_as_named_tokens, AND)
_as_named_list_selection = partial(_as_named_tokens, SELECTION)

_selection = Forward()

_factor = (_segid_factor | _residue_factor | _atom_factor | _selection)(FACTOR)

_term = Group(_factor + ZeroOrMore(Suppress(_and_op) + _factor))(TERM)

_selection_expression = Group(_term + ZeroOrMore(Suppress(_or_op) + _term))(
    SELECTION_EXPRESSION
)

_selection << (_l_paren + _selection_expression + _r_paren)(SELECTION)

_dihedral_restraint = Group(
    Suppress(_expand_literal(ASSIGN))
    + _selection.set_results_name(ATOMS_1)
    + _selection.set_results_name(ATOMS_2)
    + _selection.set_results_name(ATOMS_3)
    + _selection.set_results_name(ATOMS_4)
    + ppc.number.set_results_name(ENERGY_CONSTANT).setParseAction(ppc.convertToFloat)
    + ppc.number.set_results_name(ANGLE).setParseAction(ppc.convertToFloat)
    + ppc.number.set_results_name(RANGE).setParseAction(ppc.convertToFloat)
    + ppc.integer.set_results_name(EXPONENT).setParseAction(ppc.convertToInteger)
)
_dihedral_restraints = OneOrMore(_dihedral_restraint)

_distance_restraint = Group(
    Suppress(_expand_literal(ASSIGN))
    + _selection.set_results_name(ATOMS_1)
    + _selection.set_results_name(ATOMS_2)
    + ppc.number.set_results_name(DISTANCE).setParseAction(ppc.convertToFloat)
    + ppc.number.set_results_name(DISTANCE_MINUS).setParseAction(ppc.convertToFloat)
    + ppc.number.set_results_name(DISTANCE_PLUS).setParseAction(ppc.convertToFloat)
)

_distance_restraints = OneOrMore(_distance_restraint)


def _remove_xplor_comments(data_lines: List[str]) -> List[str]:
    result = []
    for data_line in data_lines:
        if XPLOR_COMMENT_TOKEN in data_line:
            comment_start = data_line.index(XPLOR_COMMENT_TOKEN)
            new_line = data_line[:comment_start]
            result.append(new_line)
        else:
            result.append(data_line)
    return result


def _get_single_atom_selection(
    parsed: ParseResults,
    residue_types: Dict[Tuple[str, str], str],
    default_chain_code: str = "A",
) -> AtomLabel:
    result = None
    valid = True
    invalid_reason = ""

    selections = _parse_result_to_atom_selections(parsed)

    if len(selections) != 1:
        valid = False
        invalid_reason = f"there should only be one selection i got {selections}"
    else:
        selection = selections[0]

    if valid:
        sequence_code = selection.get("resids", None)
        if sequence_code is None:
            valid = False
            invalid_reason = f"there is no sequence code in {selection}"
        elif len(sequence_code) != 1:
            valid = False
            invalid_reason = f"there is more than one sequence code in {selection}"
        else:
            sequence_code = sequence_code[0].token

    if valid:
        atom_name = selection.get("atoms", None)
        if atom_name is None:
            valid = False
            invalid_reason = f"there is no atom name in {selection}"
        elif len(atom_name) != 1:
            valid = False
            invalid_reason = f"there is more than one atom name in {selection}"
        else:
            atom_name = atom_name[0].token

    if valid:
        chain_code = selection.get("segids", None)
        if chain_code is None:
            chain_code = default_chain_code
        elif len(chain_code) > 1:
            valid = False
            invalid_reason = f"there is more than one chain code in {selection}"
        elif len(chain_code) == 1:
            chain_code = chain_code[0].token

    if valid:
        residue_type_key = chain_code, str(sequence_code)
        if residue_type_key in residue_types:
            residue_type = residue_types[residue_type_key]
        else:
            valid = False
            residue_type = None
            invalid_reason = f"the residue type for chain code: {chain_code} sequence code {sequence_code} is not known"

    if valid:
        residue = SequenceResidue(chain_code, sequence_code, residue_type)
        result = AtomLabel(residue, atom_name)

    if not valid:
        raise XPLORParseException(
            f"Couldn't get a single restraint from {parsed} because {invalid_reason}"
        )

    return result


def read_distance_restraints_or_exit_error(
    file_path: Path,
    residue_name_lookup: Dict[Tuple[str, str], str],
    chain_code: str,
    use_chains: bool = False,
) -> List[DihedralRestraint]:
    """
    read a list of dihedral restraints from a file or stream or exit

    :param file_path: the path of the file Path('-') indicates stdin
    :param residue_name_lookup:  a dictionary of residue names keyes on chain_code, residue_code
    :param chain_code: a chain code to use if no chain code is specified or use_chains is True
    :param use_chains: use the passed in chain_code rather than any read segids
    :return: a list of dihedral restraints
    """

    restraint_text = read_from_file_or_exit(
        file_path, f"xplor distance restraints from {file_path}"
    )

    file_path_display_name = get_display_file_name(file_path)

    restraints = parse_distance_restraints(
        restraint_text,
        residue_name_lookup,
        file_path_display_name,
        chain_code,
        use_chains,
    )

    return restraints


def distance_restraints_to_nef(
    restraints: List[DistanceRestraint], frame_name: str
) -> Saveframe:
    """
    convert a list of DistanceRestraints to a new distance restraint save frame
    :param restraints: the restraints
    :param frame_name: the name of the new save frame
    :return: a new save frame
    """
    loop = Loop.from_scratch(category="nef_distance_restraint")

    for tag in DISTANCE_RESTRAINT_TAGS:
        loop.add_tag(tag)

    for i, restraint in enumerate(restraints, start=1):
        selection_1 = restraint.atom_list_1[0]
        selection_2 = restraint.atom_list_2[0]
        d = restraint.target_distance
        d_minus = restraint.distance_minus
        d_plus = restraint.distance_plus

        row = {
            INDEX: i,
            RESTRAINT_ID: i,
            RESTRAINT_COMBINATION_ID: UNUSED,
            CHAIN_CODE_1: selection_1.residue.chain_code,
            SEQUENCE_CODE_1: selection_1.residue.sequence_code,
            ATOM_1: selection_1.atom_name,
            CHAIN_CODE_2: selection_2.residue.chain_code,
            SEQUENCE_CODE_2: selection_2.residue.sequence_code,
            ATOM_2: selection_2.atom_name,
            WEIGHT: 1.0,
            TARGET_VALUE: round(d, DEFAULT_PRECISION),
            LOWER_LIMIT: round(d - d_minus, DEFAULT_PRECISION),
            UPPER_LIMIT: round(d + d_plus, DEFAULT_PRECISION),
        }

        loop.add_data([row])

    save_frame_name = f"nef_distance_restraint_list_{frame_name}"

    save_frame = Saveframe.from_scratch(save_frame_name, "nef_distance_restraint")

    save_frame.add_tag("sf_category", "nef_distance_restraint_list")
    save_frame.add_tag("sf_framecode", save_frame_name)
    save_frame.add_tag("potential_type", PotentialTypes.UNDEFINED)

    save_frame.add_loop(loop)

    return save_frame


def dihedral_restraints_to_nef(
    restraints: List[DihedralRestraint], frame_name: str
) -> Saveframe:
    """
    convert a list of DihedralRestraints to a new dihedral restraint save frame
    :param restraints: the restraints
    :param frame_name: the name of the new save frame
    :return: a new save frame
    """
    loop = Loop.from_scratch(category=NEF_DIHEDRAL_RESTRAINT)

    for tag in DIHEDRAL_RESTRAINT_TAGS:
        loop.add_tag(tag)

    for i, restraint in enumerate(restraints, start=1):
        selection_1 = restraint.atom_1
        selection_2 = restraint.atom_2
        selection_3 = restraint.atom_3
        selection_4 = restraint.atom_4
        target = restraint.target_angle
        lower_limit = restraint.lower_limit
        upper_limit = restraint.upper_limit

        row = {
            INDEX: i,
            RESTRAINT_ID: i,
            RESTRAINT_COMBINATION_ID: UNUSED,
            CHAIN_CODE_1: selection_1.residue.chain_code,
            SEQUENCE_CODE_1: selection_1.residue.sequence_code,
            ATOM_1: selection_1.atom_name,
            CHAIN_CODE_2: selection_2.residue.chain_code,
            SEQUENCE_CODE_2: selection_2.residue.sequence_code,
            ATOM_2: selection_2.atom_name,
            CHAIN_CODE_3: selection_3.residue.chain_code,
            SEQUENCE_CODE_3: selection_3.residue.sequence_code,
            ATOM_3: selection_3.atom_name,
            CHAIN_CODE_4: selection_4.residue.chain_code,
            SEQUENCE_CODE_4: selection_4.residue.sequence_code,
            ATOM_4: selection_4.atom_name,
            WEIGHT: 1.0,
            TARGET_VALUE: round(target, DEFAULT_PRECISION),
            LOWER_LIMIT: round(lower_limit, DEFAULT_PRECISION),
            UPPER_LIMIT: round(upper_limit, DEFAULT_PRECISION),
        }

        loop.add_data([row])

    save_frame_name = f"nef_dihedral_restraint_list_{frame_name}"

    save_frame = Saveframe.from_scratch(save_frame_name, "nef_dihedral_restraint")

    save_frame.add_tag("sf_category", "nef_dihedral_restraint_list")
    save_frame.add_tag("sf_framecode", save_frame_name)
    save_frame.add_tag("potential_type", PotentialTypes.UNDEFINED)

    save_frame.add_loop(loop)

    return save_frame


def _get_approximate_restraint_strings(text: str) -> List[str]:

    new_lines = []
    for line in text.split("\n"):
        if XPLOR_COMMENT_TOKEN in line:
            new_line = line.split(XPLOR_COMMENT_TOKEN)[0]
            if len(new_line.strip()) > 0:
                new_lines.append(new_line)
        else:
            new_lines.append(line)

    new_lines = "\n".join(new_lines)

    restraints = new_lines.split(ASSIGN)

    if restraints[0] == "":
        restraints = restraints[1:]

    return [f"{ASSIGN} {restraint}" for restraint in restraints]


def read_dihedral_restraints_or_exit_error(
    file_path: Path,
    residue_name_lookup: Dict[Tuple[str, str], str],
    chain_code: str,
    use_chains: bool = False,
) -> List[DihedralRestraint]:
    """
    read a list of dihedral restraints from a file or stream or exit

    :param file_path: the poth of the file Path('-') indicates stdin
    :param residue_name_lookup:  a dictionary of residue names keyes on chain_code, residue_code
    :param chain_code: a chain code to use if no chain code is specified or use_chains is True
    :param use_chains: use the passed in chain_code rather than any read segids
    :return: a list of dihedral restraints
    """

    restraint_text = read_from_file_or_exit(
        file_path, f"xplor dihedral restraints from {file_path}"
    )

    file_path_display_name = get_display_file_name(file_path)

    restraints = parse_dihedral_restraints(
        restraint_text,
        residue_name_lookup,
        file_path_display_name,
        chain_code,
        use_chains,
    )

    return restraints


def parse_dihedral_restraints(
    restraint_text: str,
    residue_name_lookup: Dict[Tuple[str, str], str],
    file_path_display_name: str,
    chain_code: str,
    use_chains: bool = False,
) -> List[DihedralRestraint]:
    """
    parse xplor dihedral restraints into DihedralRestraint structures

    :param restraint_text: the text of the restraints in xplor format
    :param residue_name_lookup: a lookup for residue names from a  chain_code, residue_code key
    :param file_path_display_name: the source of the restraints for error reporting
    :param chain_code: a chain code to use for the restraints if non is provided or use_chains is true
    :param use_chains: use the passed in chain_code rather than any read segids
    :return:  a list of dihedral restraints
    """
    try:
        xplor_basic_restraints = _dihedral_restraints.ignore(XPLOR_COMMENT).parseString(
            restraint_text
        )
    except ParseException as parse_exception:
        msg = f"""\
            failed to read dihedral restraints from the file {file_path_display_name} because:
            {str(parse_exception)}
        """
        exit_error(msg)

    restraints = []
    for i, restraint in enumerate(xplor_basic_restraints, start=1):

        atom_selections = []

        for atom_index in range(1, 5):

            xplor_atoms = restraint.get(f"atoms_{atom_index}")[0]

            try:
                nef_atoms = _get_single_atom_selection(
                    xplor_atoms, residue_name_lookup, chain_code
                )

            except XPLORParseException as e:
                atom_number = end_with_ordinal(atom_index)
                approximate_restraints = _get_approximate_restraint_strings(
                    restraint_text
                )
                approximate_restraint = approximate_restraints[i - 1]
                approximate_restraint = approximate_restraint.split("\n")
                msg = f"""\
                    got a multi atom selection for the {atom_number} atom in restraint number {i}
                    in {file_path_display_name}
                    dihedral restrainst require single atom selections...
                    the restraint text is most probably:
                """
                msg = dedent(msg)
                for elem in approximate_restraint:
                    msg += f"    {elem}\n"
                exit_error(msg, e)

            atom_selections.append(nef_atoms)

        if use_chains and chain_code != ANY_CHAIN:
            atom_selections = replace_chain_in_atom_labels(atom_selections, chain_code)

        target_angle = restraint.get(ANGLE)
        angle_range = restraint.get(RANGE)
        lower_limit = target_angle - angle_range
        upper_limit = target_angle + angle_range

        restraint = DihedralRestraint(
            *atom_selections,
            target_angle=target_angle,
            lower_limit=lower_limit,
            upper_limit=upper_limit,
        )

        restraints.append(restraint)
    return restraints


def parse_distance_restraints(
    restraint_text: str,
    residue_name_lookup: Dict[Tuple[str, str], str],
    file_path_display_name: str,
    chain_code: str,
    use_chains: bool = False,
) -> List[DihedralRestraint]:
    """
    parse xplor distance restraints into DistanceRestraint structures

    :param restraint_text: the text of the restraints in xplor format
    :param residue_name_lookup: a lookup for residue names from a  chain_code, residue_code key
    :param file_path_display_name: the source of the restraints for error reporting
    :param chain_code: a chain code to use for the restraints if non is provided or use_chains is true
    :param use_chains: use the passed in chain_code rather than any read segids
    :return:  a list of dihedral restraints
    """
    try:
        xplor_basic_restraints = _distance_restraints.ignore(XPLOR_COMMENT).parseString(
            restraint_text
        )
    except ParseException as parse_exception:
        msg = f"""\
            failed to read distance restraints from the file {file_path_display_name} because:
            {str(parse_exception)}
        """
        exit_error(msg)

    restraints = []
    for i, restraint in enumerate(xplor_basic_restraints, start=1):

        atom_selections = []

        for atom_index in range(1, 3):

            xplor_atoms = restraint.get(f"atoms_{atom_index}")[0]

            try:
                nef_atoms = _get_single_atom_selection(
                    xplor_atoms, residue_name_lookup, chain_code
                )

            except XPLORParseException as e:
                atom_number = end_with_ordinal(atom_index)
                approximate_restraints = _get_approximate_restraint_strings(
                    restraint_text
                )
                approximate_restraint = approximate_restraints[i - 1]
                approximate_restraint = approximate_restraint.split("\n")
                msg = f"""\
                    got a multi atom selection for the {atom_number} atom in restraint number {i}
                    in {file_path_display_name}
                    dihedral restrainst require single atom selections...
                    the restraint text is most probably:
                """
                msg = dedent(msg)
                for elem in approximate_restraint:
                    msg += f"    {elem}\n"
                exit_error(msg, e)

            atom_selections.append(nef_atoms)

        if use_chains and chain_code != ANY_CHAIN:
            atom_selections = replace_chain_in_atom_labels(atom_selections, chain_code)

        target_distance = restraint.get(DISTANCE)
        distance_minus = restraint.get(DISTANCE_MINUS)
        distance_plus = restraint.get(DISTANCE_PLUS)

        atom_selections = [[atom_selection] for atom_selection in atom_selections]
        restraint = DistanceRestraint(
            *atom_selections,
            target_distance=target_distance,
            distance_minus=target_distance - distance_minus,
            distance_plus=target_distance + distance_plus,
        )

        restraints.append(restraint)

    return restraints


#
#
# if __name__ == "__main__":
#
#     # with open('src/nef_pipelines/tests/xplor/test_data/noes_S135A_BBBB_AAAA.tbl') as fh:
#     # # with open('src/nef_pipelines/tests/xplor/test_data/noes_S135A_AAAA_BBBB.tbl') as fh:
#     # # with open('src/nef_pipelines/tests/xplor/test_data/noes_S135A_AAAA.tbl') as fh:
#     #     lines = fh.readlines()
#     #
#     #     lines = remove_xplor_comments(lines)
#     #     lines = ''.join(lines)
#     #
#     #     xplor_basic_restraints = distance_restraints.parseString(lines)
#     #
#     #     restraints = []
#     #     for restraint in xplor_basic_restraints:
#     #         atoms_1 = restraint.get('atoms_1')
#     #         atoms_2 = restraint.get('atoms_2')
#     #         d = restraint.get('d')
#     #         d_minus = restraint.get('d_minus')
#     #         d_plus = restraint.get('d_plus')
#     #
#     #         residue_types = {('AAAA', i): 'ALA' for i in range(1, 1000)}
#     #         residue_types.update({('BBBB', i): 'ALA' for i in range(1, 1000)})
#     #
#     #         selection_1 = get_single_atom_selection(atoms_1, residue_types)
#     #         selection_2 = get_single_atom_selection(atoms_2, residue_types)
#     #
#     #         distance_restraint = DistanceRestraint(atom_list_1 =[selection_1], atom_list_2 =[selection_2],
#                                                      distance = d, distance_minus=d_minus, distance_plus=d_plus)
#     #
#     #         restraints.append(distance_restraint)
#     #
#     #     offset = 44
#     #     for restraint in restraints:
#     #         print(restraint)
#     #         selection_1 = restraint.atom_list_1[0]
#     #         selection_2 = restraint.atom_list_2[0]
#     #
#     #         if selection_1.residue.chain_code == 'AAAA':
#     #             residue_1 = replace(selection_1.residue, sequence_code=selection_1.residue.sequence_code + offset)
#     #             selection_1 = replace(selection_1, residue=residue_1)
#     #
#     #         if selection_2.residue.chain_code == 'AAAA':
#     #             residue_2 = replace(selection_2.residue, sequence_code=selection_2.residue.sequence_code + offset)
#     #             selection_2 = replace(selection_2, residue=residue_2)
#     #
#     #         restraint.atom_list_1[0] = selection_1
#     #         restraint.atom_list_2[0] = selection_2
#     #
#     #     loop = restraints_to_nef(restraints)
#     # print(loop)
#     #
#     # rows = []
#     # for nef_row in loop_row_namespace_iter(loop, convert=False):
#     #     segid_1 = nef_row.chain_code_1
#     #     resid_1 = nef_row.sequence_code_1
#     #     name_1 = nef_row.atom_name_1
#     #     selection_1 = '(', 'segid',  segid_1,  'and',  'resid',  resid_1, 'and', 'name',  name_1, ')'
#     #
#     #     segid_2 = nef_row.chain_code_2
#     #     resid_2 = nef_row.sequence_code_2
#     #     name_2 = nef_row.atom_name_2
#     #     selection_2 = '(', 'segid',  segid_2,  'and',  'resid',  resid_2, 'and', 'name',  name_2, ')'
#     #
#     #     d = nef_row.target_value
#     #     d_minus = d - nef_row.lower_limit
#     #     d_plus = nef_row.upper_limit - d
#     #
#     #     row = ['assign', *selection_1, *selection_2, d, d_minus, d_plus]
#     #
#     #     rows.append(row)
#     #
#     # table = tabulate(rows, tablefmt='plain', floatfmt="7.3f")
#     # table = table.replace('assign  ', 'assign ')
#     # table = table.replace('(  ', '(')
#     # table = table.replace('  )', ')')
#     # table = table.replace('   )', ')   ')
#     # table = table.replace('  )', ')  ')
#     # table = table.replace(' )', ') ')
#     # table = table.replace('  and  ', ' and ')
#     # table = table.replace('segid  ', 'segid ')
#     # table = table.replace('resid  ', 'resid ')
#     # table = table.replace('name  ', 'name ')
#     #
#     # print(table)
#
#     with open("src/nef_pipelines/tests/xplor/test_data/dihedrals_S135A_AAAA.tbl") as fh:
#         # with open('src/nef_pipelines/tests/xplor/test_data/noes_S135A_AAAA_BBBB.tbl') as fh:
#         # with open('src/nef_pipelines/tests/xplor/test_data/noes_S135A_AAAA.tbl') as fh:
#         lines = fh.readlines()
#
#         lines = remove_xplor_comments(lines)
#         lines = "".join(lines)
#
#         xplor_basic_restraints = dihedral_restraints.parseString(lines)
#
#         restraints = []
#
#         for restraint in xplor_basic_restraints:
#             atoms_1 = restraint.get("atoms_1")
#             atoms_2 = restraint.get("atoms_2")
#             atoms_3 = restraint.get("atoms_3")
#             atoms_4 = restraint.get("atoms_4")
#
#             target_angle = restraint.get("angle")
#             angle_range = restraint.get("range")
#             lower_limit = target_angle - angle_range
#             upper_limit = target_angle + angle_range
#
#             residue_types = {("AAAA", i): "ALA" for i in range(1, 1000)}
#             residue_types.update({("BBBB", i): "ALA" for i in range(1, 1000)})
#
#             selection_1 = get_single_atom_selection(atoms_1, residue_types)
#             selection_2 = get_single_atom_selection(atoms_2, residue_types)
#             selection_3 = get_single_atom_selection(atoms_3, residue_types)
#             selection_4 = get_single_atom_selection(atoms_4, residue_types)
#
#             dihedral_restraint = DihedralRestraint(
#                 atom_list_1=[selection_1],
#                 atom_list_2=[selection_2],
#                 atom_list_3=[selection_3],
#                 atom_list_4=[selection_4],
#                 target=target_angle,
#                 lower_limit=lower_limit,
#                 upper_limit=upper_limit,
#             )
#
#             restraints.append(dihedral_restraint)
#
#         offset = 44
#         for restraint in restraints:
#
#             selection_1 = restraint.atom_list_1[0]
#             selection_2 = restraint.atom_list_2[0]
#             selection_3 = restraint.atom_list_3[0]
#             selection_4 = restraint.atom_list_4[0]
#
#             if selection_1.residue.chain_code == "AAAA":
#                 residue_1 = replace(
#                     selection_1.residue,
#                     sequence_code=selection_1.residue.sequence_code + offset,
#                 )
#                 selection_1 = replace(selection_1, residue=residue_1)
#
#             if selection_2.residue.chain_code == "AAAA":
#                 residue_2 = replace(
#                     selection_2.residue,
#                     sequence_code=selection_2.residue.sequence_code + offset,
#                 )
#                 selection_2 = replace(selection_2, residue=residue_2)
#
#             if selection_3.residue.chain_code == "AAAA":
#                 residue_3 = replace(
#                     selection_3.residue,
#                     sequence_code=selection_3.residue.sequence_code + offset,
#                 )
#                 selection_3 = replace(selection_3, residue=residue_3)
#
#             if selection_4.residue.chain_code == "AAAA":
#                 residue_4 = replace(
#                     selection_4.residue,
#                     sequence_code=selection_4.residue.sequence_code + offset,
#                 )
#                 selection_4 = replace(selection_4, residue=residue_4)
#
#             restraint.atom_list_1[0] = selection_1
#             restraint.atom_list_2[0] = selection_2
#             restraint.atom_list_3[0] = selection_3
#             restraint.atom_list_4[0] = selection_4
#
#         loop = dihedral_restraints_to_nef(restraints)
#     # print(loop)
#     #
#     rows = []
#     for nef_row in loop_row_namespace_iter(loop, convert=False):
#         segid_1 = nef_row.chain_code_1
#         resid_1 = nef_row.sequence_code_1
#         name_1 = nef_row.atom_name_1
#         selection_1 = (
#             "(",
#             "segid",
#             segid_1,
#             "and",
#             "resid",
#             resid_1,
#             "and",
#             "name",
#             name_1,
#             ")",
#         )
#
#         segid_2 = nef_row.chain_code_2
#         resid_2 = nef_row.sequence_code_2
#         name_2 = nef_row.atom_name_2
#         selection_2 = (
#             "(",
#             "segid",
#             segid_2,
#             "and",
#             "resid",
#             resid_2,
#             "and",
#             "name",
#             name_2,
#             ")",
#         )
#
#         segid_3 = nef_row.chain_code_3
#         resid_3 = nef_row.sequence_code_3
#         name_3 = nef_row.atom_name_3
#         selection_3 = (
#             "(",
#             "segid",
#             segid_3,
#             "and",
#             "resid",
#             resid_3,
#             "and",
#             "name",
#             name_3,
#             ")",
#         )
#
#         segid_4 = nef_row.chain_code_4
#         resid_4 = nef_row.sequence_code_4
#         name_4 = nef_row.atom_name_4
#         selection_4 = (
#             "(",
#             "segid",
#             segid_4,
#             "and",
#             "resid",
#             resid_4,
#             "and",
#             "name",
#             name_4,
#             ")",
#         )
#
#         target = nef_row.target_value
#         range_1 = round(target - nef_row.lower_limit, 5)
#         range_2 = round(nef_row.upper_limit - target, 5)
#
#         row_1 = ["assign", *selection_1, *selection_2, " ", " ", " ", " "]
#         row_2 = ["     ", *selection_3, *selection_4, 1.0, target, range_1, 2]
#
#         rows.append(row_1)
#         rows.append(row_2)
#
#     table = tabulate(rows, tablefmt="plain", floatfmt="7.3f")
#     table = table.replace("assign  ", "assign ")
#     table = table.replace("        (segid", "       (segid")
#     table = table.replace("(  ", "(")
#     table = table.replace("  )", ")")
#     table = table.replace("   )", ")   ")
#     table = table.replace("  )", ")  ")
#     table = table.replace(" )", ") ")
#     table = table.replace("  and  ", " and ")
#     table = table.replace("segid  ", "segid ")
#     table = table.replace("resid  ", "resid ")
#     table = table.replace("name  ", "name ")
#     table = table.replace("        (segid", "       (segid")
#
#     print(table)


def _drill_down_0(selection):
    result = []
    target = selection
    while isinstance(target, (ParseResults, _NamedToken)) and len(target) > 0:
        result.append(target.get_name())
        target = target[0]
    return result


def _get_selection_expressions_from_selection(
    selection: ParseResults, selection_text
) -> List[ParseResults]:
    drill_down = _drill_down_0(selection)

    result = []
    if len(drill_down) >= 4:
        if drill_down[0:3] == ["selection", "selection-expression", "term"]:

            term_names = set([term.get_name() for term in selection[0][0]])
            atom_selection_names = {"segid", "resid", "atom"}
            atom_selection_term_names = term_names.intersection(atom_selection_names)

            if (
                len(term_names) == 1
                and next(iter(term_names)) == "selection-expression"
            ):
                for elem in selection[0]:
                    result.append(elem[0])

            elif len(atom_selection_term_names) > 0:
                result.append(selection[0])
            else:
                msg = f"""\
                   selection error

                   selections must contain selections at the first or  second level, so they should look something
                   like this :
                       (<selection-1>)
                   or  ((<selection-1>) OR (<selection-2>)...) etc

                   where the terms  might be something like
                       segid ZZZZ resid 2 and name HA
                       segid XXXX resid 1 and name HN

                   i got a selection that wasn't in this format {selection_text}

                   value was: {_selection_expression}
                """
                raise XPLORParseException(msg)
    elif (
        len(drill_down) == 3
        and drill_down[0] == "selection-expression"
        and len(selection) == 1
    ):
        result = [selection]

    return result


def _get_atom_selections_from_selection_expression(selection_expression: ParseResults):

    if not selection_expression.get_name() == "selection-expression":
        msg = f"""\
            i expected a <selection-expression> but got a <{selection_expression.get_name()}>

            a selection expression should look something like

            (segid  AAAA and resid 1 and name HA)

            input value was: {selection_expression} [this is not what you entered but the parse result for debugging]

        """
        raise XPLORParseException(msg)

    result = {}
    for term in selection_expression:
        if term.get_name() == "term":
            for elem in term:
                elem_name = elem.get_name()

                if elem_name in ("segid", "resid", "atom"):
                    result.setdefault(f"{elem_name}s", []).append(elem)
                elif elem_name == "selection-expression":

                    sub_terms = _get_atom_selections_from_selection_expression(elem)

                    for sub_term_name, sub_term in sub_terms.items():
                        result[sub_term_name] = sub_term

        else:
            raise XPLORParseException(
                f"implementation error unexpected element type {elem_name} found in {selection_expression}"
            )

    return result


def _string_to_atom_selections(selection_text):
    parse_result = _selection.parseString(selection_text, parse_all=True)

    return _parse_result_to_atom_selections(parse_result)


def _parse_result_to_atom_selections(parse_result):
    selections = _get_selection_expressions_from_selection(
        parse_result, str(parse_result)
    )
    return [
        _get_atom_selections_from_selection_expression(selection_expression)
        for selection_expression in selections
    ]


def _exit_if_chains_and_filenames_dont_match(chains, file_names):
    num_file_names = len(file_names)
    num_chains = len(chains)
    if num_file_names != num_chains:
        msg = f"""\
            your provided {num_file_names} files and {num_chains} chains
            there must be a filename for each chain you provide
            file names were: {','.join(file_names)}
            chains were: {','.join(chains)}
        """
        msg = dedent(msg)
        exit_error(msg)



# original implementation by esther
import sys
from pathlib import Path
from string import digits
from typing import Iterable, Iterator, List, Tuple

import typer
from pynmrstar import Entry, Saveframe
from tabulate import tabulate

from nef_pipelines.lib.nef_lib import (
    loop_row_namespace_iter,
    read_entry_from_file_or_stdin_or_exit_error,
)
from nef_pipelines.lib.structures import LineInfo
from nef_pipelines.lib.util import STDIN, end_with_ordinal, exit_error
from nef_pipelines.transcoders.mars import export_app

NMR_RESIDUE = "nmr_residue"

NMR_CHAIN = "nmr_chain"

STDOUT_PATH = "-"

app = typer.Typer()

# TODO: name translations
# TODO: correct weights
# TODO: move utuilities to lib
# TODO: support multiple chains

REMOVE_DIGITS = str.maketrans("", "", digits)


def has_numbers(input: str) -> bool:
    return any(char.isdigit() for char in input)


# noinspection PyUnusedLocal
@export_app.command()
def fragments(
    input: Path = typer.Option(
        STDIN,
        "--input",
        "-i",
        metavar="|INPUT|",
        help="input to read NEF data from [stdin = -]",
    ),
    output_file: str = typer.Argument(
        None,
        help="file name to output to [default <entry_id>_fix_con.tab] for stdout use -",
        metavar="<MARS_SHIFT_FILE>",
    ),
):
    """- convert nef chemical shifts to mars"""

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    output_file = (
        f"{entry.entry_id}_fix_ass.tab" if output_file is None else output_file
    )

    table = pipe(entry)

    table = _format_table(table)
    file_h = sys.stdout if output_file == "-" else open(output_file, "w")

    print(tabulate(table, tablefmt="plain"), file=file_h)

    if output_file != STDOUT_PATH:
        file_h.close()

    if output_file != STDOUT_PATH:
        if not sys.stdout.isatty():
            print(entry)


def _format_table(table: List[List[Tuple[str, str]]]) -> List[List[str]]:
    max_overall_elem_length = 0
    for chain in table:
        max_elem_length = max([len(" ".join(prs)) for prs in chain])
        max_overall_elem_length = (
            max_elem_length
            if max_elem_length > max_overall_elem_length
            else max_overall_elem_length
        )
    for chain in table:
        for i, prs in enumerate(chain):
            pr_length = len(prs[0]) + len(prs[1])
            pad = " " * (max_overall_elem_length - pr_length)
            pr_pair = f"{prs[0]}{pad}{prs[1]}"
            chain[i] = pr_pair
    return table


def pipe(entry: Entry) -> List[List[Tuple[str, str]]]:
    frame = _get_assignment_frame_or_exit(entry)
    connected_residues = _get_connected_residues_and_chains(frame, entry)
    return _build_connected_pseudo_residues(connected_residues)


def _build_connected_pseudo_residues(connected_residues):
    table = []
    ends = []
    for chain_code, residue_codes in connected_residues.items():
        chain = [
            (f"PR_{first}", f"PR_{second}")
            for first, second in overlapped_pairs(residue_codes)
        ]
        ends.append(f"#{chain_code}")
        table.append(chain)

    return table


def _get_connected_residues_and_chains(frame: Saveframe, entry: Entry):
    connected_chains = set()
    for chain in loop_row_namespace_iter(frame.get_loop(NMR_CHAIN)):
        if chain.is_connected:
            connected_chains.add(chain.short_name)
    connected_residues = {}
    nmr_residue_loop = frame.get_loop(NMR_RESIDUE)
    for line_no, residue in enumerate(
        loop_row_namespace_iter(nmr_residue_loop), start=1
    ):
        chain_code = residue.chain_code

        if chain_code in connected_chains:
            line = nmr_residue_loop[line_no - 1]
            line_info = LineInfo(entry.entry_id, line_no, line)
            sequence_code = _get_connected_residue_or_exit_error(residue, line_info)

            connected_residues.setdefault(chain_code, []).append(sequence_code)
    for connected_chain, connected_residue_list in connected_residues.items():
        connected_residues[connected_chain] = remove_duplicates_stable(
            connected_residue_list
        )
    return connected_residues


def _get_assignment_frame_or_exit(entry):
    frames = entry.get_saveframes_by_category("ccpn_assignment")
    if len(frames) == 0:
        exit_error(f"no assignment frame found in entry {entry.name}")
    if len(frames) > 1:
        msg = f"""\
                    there should only be one assignment frame i found {len(frames)} in entry {entry.name}
                    there names were {' '.join([frame.name for frame in frames])}
            """
        exit_error(msg)
    frame = frames[0]
    return frame


def _get_connected_residue_or_exit_error(residue, line_info: LineInfo):
    sequence_code = residue.sequence_code
    if sequence_code.startswith("@"):
        sequence_code = sequence_code.lstrip("@")
        sequence_code_fields = sequence_code.split("-")
        if len(sequence_code_fields) == 0 or len(sequence_code_fields) > 2:
            msg = f"""\
                for the {end_with_ordinal(line_info.line_no)} line in the frame '_save_nmr_residue in frame
                save_ccpn_assignment in entity {line_info.file_name} the sequence code {sequence_code} had an unexpected
                format it should be @i or @i-1 where i is a number, the line was
                {line_info.line}
            """
            exit_error(msg)
        sequence_code = sequence_code_fields[0]
    return sequence_code


# https://stackoverflow.com/questions/480214/how-do-i-remove-duplicates-from-a-list-while-preserving-order
def remove_duplicates_stable(seq: Iterable) -> List:

    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


# https://stackoverflow.com/questions/21303224/iterate-over-all-pairs-of-consecutive-items-in-a-list
def overlapped_pairs(seq: Iterator) -> Iterator[Tuple]:
    for first, second in zip(seq, seq[1:]):
        yield first, second


def _filter_headings_by_pseudoatoms(base_headers, pseudo_residues):
    atom_names = set()
    for pseudo_residue in pseudo_residues.values():
        for pseudo_atom in pseudo_residue.values():
            offset = (
                f"-{pseudo_atom.negative_offset}"
                if pseudo_atom.negative_offset == 1
                else ""
            )
            atom_names.add(f"{pseudo_atom.atom_name.upper()}{offset}")
    headers = []
    for header in base_headers:
        if header == "" or header in atom_names:
            headers.append(header)
    return headers



from pathlib import Path
from typing import Dict, List, Tuple

import typer
from pynmrstar import Entry, Saveframe

from nef_pipelines.lib.constants import NEF_PIPELINES
from nef_pipelines.lib.nef_lib import read_entry_from_file_or_stdin_or_exit_error
from nef_pipelines.lib.sequence_lib import chain_code_iter, sequence_from_entry_or_exit
from nef_pipelines.lib.shift_lib import shifts_to_nef_frame
from nef_pipelines.lib.structures import SequenceResidue
from nef_pipelines.lib.util import (
    STDIN,
    cached_file_stream,
    exit_error,
    fixup_metadata,
    get_pipe_file_text_or_exit,
    get_version,
    script_name,
)
from nef_pipelines.transcoders.nmrview import import_app
from nef_pipelines.transcoders.nmrview.nmrview_lib import parse_shifts, read_sequence

app = typer.Typer()

# TODO: add a force operation to replace the frame


# noinspection PyUnusedLocal
@import_app.command(no_args_is_help=True)
def shifts(
    chain_codes: str = typer.Option(
        "A",
        "--chains",
        help="chain codes as a list of names separated by dots",
        metavar="<CHAIN-CODES>",
    ),
    frame_name: str = typer.Option(
        "nmrview", "-f", "--frame-name", help="a name for the frame"
    ),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--input",
        metavar="|PIPE|",
        help="input to read NEF data from [- is stdin]",
    ),
    file_names: List[Path] = typer.Argument(
        ..., help="input files of type nmrview.out", metavar="<NMRVIEW-shifts>.out"
    ),
):
    """convert nmrview shift file <nmrview-shifts>.out to NEF"""

    try:
        entry = read_entry_from_file_or_stdin_or_exit_error(input)

        sequence = sequence_from_entry_or_exit(entry)

        pipe(entry, chain_codes, sequence, frame_name, file_names)

    except Exception as e:
        exit_error(f"reading sequence failed because {e}", e)


def pipe(entry, chain_codes, sequence, entry_name, file_names):

    nmrview_frames = []

    for file_name, chain_code in zip(file_names, chain_code_iter(chain_codes)):

        with cached_file_stream(file_name) as lines:

            chain_seqid_to_type = _sequence_to_residue_type_lookup(sequence)

            nmrview_shifts = parse_shifts(
                lines, chain_seqid_to_type, chain_code=chain_code, file_name=file_name
            )

            frame = shifts_to_nef_frame(nmrview_shifts, entry_name)

            nmrview_frames.append(frame)

    entry = add_frames_to_entry(entry, nmrview_frames)

    print(entry)


def add_frames_to_entry(entry: Entry, frames: List[Saveframe]) -> Entry:
    # TODO deal with merging esp wrt to molecular systems and possibly with other information
    # TODO add frame rename and frame delete
    """
    take a set of save frames and  add them to an Entry and update the NEF metadata header

    Args:
        entry: an entry to add the save frames to
        frames: a set of save frames to add, they must have different names to those present already


    Returns:
        the updated entry containing the frames
    """

    fixup_metadata(entry, NEF_PIPELINES, get_version(), script_name(__file__))

    for frame in frames:

        new_frame_name = frame.name

        frame_in_entry = save_frame_name_in_entry(entry, new_frame_name)

        if frame_in_entry:
            msg = (
                f"the frame named {new_frame_name} already exists in the stream, rename it or delete to add "
                f"the new frame shift frame"
            )
            exit_error(msg)

        entry.add_saveframe(frame)

    return entry


def save_frame_name_in_entry(entry, new_frame_name):
    frame_in_entry = False
    try:
        entry.get_saveframe_by_name(new_frame_name)
    except KeyError:
        frame_in_entry = False

    return frame_in_entry


def sequence_from_frames(frames: Saveframe) -> List[SequenceResidue]:

    residues = []
    for frame in frames:
        for loop in frame:
            chain_code_index = loop.tag_index("chain_code")
            sequence_code_index = loop.tag_index("sequence_code")
            residue_name_index = loop.tag_index("residue_name")

            for line in loop:
                chain_code = line[chain_code_index]
                sequence_code = int(line[sequence_code_index])
                residue_name = line[residue_name_index]
                residue = SequenceResidue(chain_code, sequence_code, residue_name)
                residues.append(residue)

    return residues


# TODO this should be replaced by a library function from nef or sequence utils...
# also need ro report what file we are trying to read shifts from
def _get_sequence_or_exit(args):
    sequence_file = None
    if "sequence" in args:
        sequence_file = args.sequence

    sequence = None
    if not sequence_file:
        try:
            lines = get_pipe_file_text_or_exit(args)

            entry = Entry.from_string(lines)
            frames = entry.get_saveframes_by_category("nef_molecular_system")
            sequence = sequence_from_frames(frames)

        except Exception as e:
            exit_error(f"failed to read sequence from input stream because {e}", e)

    else:
        with open(sequence_file, "r") as lines:
            sequence = read_sequence(lines, chain_code=args.chain_code)
    return sequence


def _sequence_to_residue_type_lookup(
    sequence: List[SequenceResidue],
) -> Dict[Tuple[str, int], str]:
    result: Dict[Tuple[str, int], str] = {}
    for residue in sequence:
        result[residue.chain_code, residue.sequence_code] = residue.residue_name
    return result


if __name__ == "__main__":

    typer.run(shifts)



import typer

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.sparky.importers.shifts import shifts

SHIFTS_SPARKY = "nef_chemical_shift_list_sparky"

app = typer.Typer()
app.command()(shifts)

EXPECTED = """\
save_nef_chemical_shift_list_sparky
    _nef_chemical_shift_list.sf_category   nef_chemical_shift_list
    _nef_chemical_shift_list.sf_framecode  nef_chemical_shift_list_sparky


    loop_
        _nef_chemical_shift.chain_code
        _nef_chemical_shift.sequence_code
        _nef_chemical_shift.residue_name
        _nef_chemical_shift.atom_name
        _nef_chemical_shift.value
        _nef_chemical_shift.value_uncertainty
        _nef_chemical_shift.element
        _nef_chemical_shift.isotope_number

        A   236   PRO   C      176.443   .   C   13
        A   236   PRO   CA     62.822    .   C   13
        A   236   PRO   CB     32.269    .   C   13
        A   236   PRO   CD     49.668    .   C   13
        A   236   PRO   CG     27.035    .   C   13
        A   236   PRO   HA     4.438     .   H   1
        A   236   PRO   HB2    1.918     .   H   1
        A   236   PRO   HB3    2.271     .   H   1
        A   236   PRO   HD2    3.541     .   H   1
        A   236   PRO   HG2    1.967     .   H   1
        A   237   ALA   C      177.765   .   C   13
        A   237   ALA   CA     52.387    .   C   13
        A   237   ALA   CB     19.145    .   C   13
        A   237   ALA   HN     8.549     .   H   1
        A   237   ALA   HA     4.277     .   H   1
        A   237   ALA   N      124.638   .   N   15
        A   237   ALA   QB     1.357     .   H   1
        A   238   MET   C      175.895   .   C   13
        A   238   MET   CA     55.484    .   C   13
        A   238   MET   CB     32.904    .   C   13
        A   238   MET   CG     31.928    .   C   13
        A   238   MET   HN     8.379     .   H   1
        A   238   MET   HA     4.412     .   H   1
        A   238   MET   HB2    1.956     .   H   1
        A   238   MET   HB3    2.018     .   H   1
        A   238   MET   HG2    2.505     .   H   1
        A   238   MET   HG3    2.565     .   H   1
        A   238   MET   N      119.821   .   N   15
    stop_
save_
"""


def test_ppm_out_short():

    STREAM = open(path_in_test_data(__file__, "P3a_L273R_sequence_short.neff")).read()

    path = path_in_test_data(__file__, "test_shifts_P3a_L273R_shifts_short.txt")

    result = run_and_report(app, [path], input=STREAM)

    shifts_result = isolate_frame(result.stdout, SHIFTS_SPARKY)

    assert_lines_match(EXPECTED, shifts_result)



import string
from collections import Counter
from dataclasses import replace
from pathlib import Path
from textwrap import dedent
from typing import Dict, Iterable, List, Optional, Tuple, Union

from ordered_set import OrderedSet
from pynmrstar import Entry, Loop, Saveframe

from nef_pipelines.lib.constants import NEF_UNKNOWN
from nef_pipelines.lib.nef_lib import UNUSED, loop_row_namespace_iter

# from nef_pipelines.lib.nef_lib import loop_to_dataframe
from nef_pipelines.lib.structures import AtomLabel, Linking, SequenceResidue
from nef_pipelines.lib.util import (
    chunks,
    exit_error,
    get_display_file_name,
    is_int,
    read_from_file_or_exit,
)

NEF_CHAIN_CODE = "chain_code"
ANY_CHAIN = "_"


def chain_code_iter(
    user_chain_codes: Union[str, List[str]] = "", exclude: Union[str, List[str]] = ()
) -> Iterable[str]:
    """
    split input string into chain codes separated by .s, and yield them.
    Then yield any remaining letters of the upper case alphabet till they run out

    Args:
        user_chain_codes (str):  string of dot separated chain codes
        exclude: chain codes to not include in the iteration

    Returns:
        Iterable[str] single codes
    """

    ascii_uppercase = list(string.ascii_uppercase)

    if isinstance(user_chain_codes, str):
        if "." in user_chain_codes:
            user_chain_codes = user_chain_codes.split(".")

    if isinstance(exclude, str):
        if "." in exclude:
            exclude = exclude.split(".")

    chain_codes = user_chain_codes if user_chain_codes else ascii_uppercase

    seen_codes = set()

    for chain_code in chain_codes:
        seen_codes.add(chain_code)
        if chain_code not in exclude:
            yield chain_code

    for chain_code in ascii_uppercase:
        if chain_code not in seen_codes and chain_code not in exclude:
            yield chain_code

    raise ValueError("run out of chain names!")


def get_linking(
    target_sequence: List[SequenceResidue],
    no_chain_start=List[str],
    no_chain_end=List[str],
) -> List[SequenceResidue]:

    """
    get the correct linking for residue in a standard sequenced chain

    Args:
        target_sequence (List[SequenceResidue]) : the sequence
        no_chain_start (List[str]): if the chain isn't in the list cap it at the start
        no_end (List[str]): if the chain isn't in the list cap it at the end

    Returns:
         List[SequenceResidue]: correctly linked residues
    """

    by_chain = {}
    for residue in target_sequence:
        by_chain.setdefault(residue.chain_code, []).append((residue, "middle"))

    result = []
    for chain, residues_and_linkages in by_chain.items():

        if chain not in no_chain_start:
            residues_and_linkages[0] = (residues_and_linkages[0][0], "start")

        if chain not in no_chain_end:
            residues_and_linkages[-1] = (residues_and_linkages[-1][0], "end")

        result.extend(residues_and_linkages)

    return result


def sequence_to_nef_frame(
    sequences: List[SequenceResidue],
    no_chain_start: Union[List[str], Tuple[str]] = (),
    no_chain_end: Union[List[str], Tuple[str]] = (),
) -> Saveframe:
    """

    Args:
        sequences (List[SequenceResidue]): the sequence as a list of sequence residues [will be sorted]
        no_chain_start (List[str]): if the chain isn't in the list cap it at the start
        no_chain_end (List[str]): if the chain isn't in the list cap it at the end

    Returns:
        Saveframe: a NEF saveframe
    """

    sequences = sorted(sequences)

    category = "nef_molecular_system"

    frame_code = f"{category}"

    nef_frame = Saveframe.from_scratch(frame_code, category)

    nef_frame.add_tag("sf_category", category)
    nef_frame.add_tag("sf_framecode", frame_code)

    nef_loop = Loop.from_scratch("nef_sequence")
    nef_frame.add_loop(nef_loop)

    tags = (
        "index",
        "chain_code",
        "sequence_code",
        "residue_name",
        "linking",
        "residue_variant",
        "cis_peptide",
    )

    nef_loop.add_tag(tags)

    # TODO need tool to set ionisation correctly
    residue_and_linkages = get_linking(sequences, no_chain_start, no_chain_end)

    for index, (sequence_residue, linking) in enumerate(residue_and_linkages):

        nef_loop.add_data_by_tag("index", index + 1)
        nef_loop.add_data_by_tag("chain_code", sequence_residue.chain_code)
        nef_loop.add_data_by_tag("sequence_code", sequence_residue.sequence_code)
        nef_loop.add_data_by_tag("residue_name", sequence_residue.residue_name.upper())
        nef_loop.add_data_by_tag("linking", linking)
        nef_loop.add_data_by_tag("residue_variant", NEF_UNKNOWN)
        nef_loop.add_data_by_tag("cis_peptide", NEF_UNKNOWN)

    return nef_frame


TRANSLATIONS_3_1 = {
    "ALA": "A",
    "ARG": "R",
    "ASN": "N",
    "ASP": "D",
    "CYS": "C",
    "GLU": "E",
    "GLN": "Q",
    "GLY": "G",
    "HIS": "H",
    "ILE": "I",
    "LEU": "L",
    "LYS": "K",
    "MET": "M",
    "PHE": "F",
    "PRO": "P",
    "SER": "S",
    "THR": "T",
    "TRP": "W",
    "TYR": "Y",
    "VAL": "V",
}
TRANSLATIONS_1_3 = {value: key for (key, value) in TRANSLATIONS_3_1.items()}


class BadResidue(Exception):
    """
    Bad residue found in a sequence
    """

    pass


def translate_1_to_3(
    sequence: str,
    translations: Dict[str, str] = TRANSLATIONS_1_3,
    unknown: Optional[str] = None,
) -> List[str]:
    """
    Translate a 1 letter sequence to a 3 letter sequence
    Args:
        sequence (str): 1 letter sequence
        translations (Dict[str, str]): a list of translations for single amino acid codes to 3 letter residue names
        unknown (Optional[str]): optional name for residues if they are unknown, if set no error is raised if a
                                 1 letter residue name is not recognised
    Returns List[str]:
        a list of 3 residue codes

    """
    result = []
    for i, residue_name_1let in enumerate(sequence):
        residue_name_1let = residue_name_1let.upper()
        if residue_name_1let in translations:
            result.append(translations[residue_name_1let])
        else:
            if unknown:
                result.append(unknown)
            else:
                msg = f"""\
                     unknown residue {residue_name_1let} at residue {i+1}
                     sequence: {sequence}
                               {(' ' * i) + '^'}
                """
                msg = dedent(msg)
                raise BadResidue(msg)

    return result


def translate_3_to_1(
    sequence: List[str], translations: Dict[str, str] = TRANSLATIONS_3_1
) -> List[str]:
    """

    Translate a 3 letter sequence to a 1 letter sequence
    Args:
        sequence (str): 3 letter sequence
        translations (Dict[str, str]): a list of translations for single amino acid codes to 3 letter residue names

    Returns List[str]:
        a list of 1 residue codes

    :param sequence:
    :return:
    """
    result = []
    for i, resn in enumerate(sequence, start=1):
        resn = resn.upper()
        if resn in translations:
            result.append(translations[resn])
        else:
            msg = f"""
            unknown residue {resn}
            in sequence {chunks(' '.join(sequence), 10)}
            at residue number {i}
            """
            exit_error(msg)

    return result


def sequence_3let_to_sequence_residues(
    sequence_3let: List[str], chain_code: str = "A"
) -> List[SequenceResidue]:
    """
    Translate a list of 3 residue sequence codes to SequenceResidues
    Args:
        sequence_3let (List([str]): list of 3 letter residue names
        chain_code (str): the chain code [defaults to A]
        offset (int): the sequence offset [defaults to 0]

    Returns List[SequenceResidue]:
        the sequence as a list of SequenceResidues

    """

    return [
        SequenceResidue(chain_code, i, residue)
        for (i, residue) in enumerate(sequence_3let, start=1)
    ]


def sequence_residues_to_sequence_3let(
    sequence: List[SequenceResidue], chain_code: str = "A"
) -> List[str]:
    """
    Translate a list of SequenceResidues to 3 letter sequence codes t
    Args:
        sequence (List([SequenceResidues]): list of residues
        chain_code (str): the chain code [defaults to A] only residues fromthis chain will be converted

    Returns List[str]:
        the sequence as a list of 3 letter residues

    """
    return [
        residue.residue_name for residue in sequence if residue.chain_code == chain_code
    ]


def frame_to_chains(sequence_frame: Saveframe) -> List[str]:
    """
     a nef molecular systems list the chains found

    :param sequence_frames: a nef molecular system save frame
    :return: a list of chain_codes
    """

    chains = set()
    for loop in sequence_frame.loop_dict.values():
        for row in loop_row_namespace_iter(loop):
            chains.add(row.chain_code)

    if NEF_UNKNOWN in chains:
        chains.remove(NEF_UNKNOWN)

    chains = sorted(chains)

    return chains


def count_residues(sequence_frame: Saveframe, chain_code: str) -> Dict[str, int]:
    """
    given a nef molecular system and a chain list the number of residues of each type in the chain

    :param sequence_frames: a nef molecular system save frame
    :return: a list of chain_codes
    """

    residue_number_to_residue_name = {}
    for row in loop_row_namespace_iter(sequence_frame.loops[0]):
        if row.chain_code == chain_code:
            residue_number_to_residue_name[row.sequence_code] = row.residue_name

    result = Counter()
    for residue_name in sorted(residue_number_to_residue_name.values()):
        result[residue_name] += 1

    return result


def get_sequence_or_exit(
    file_name: Path = Path("-"),
) -> Optional[List[SequenceResidue]]:
    """
    read a sequence from a nef molecular system on a file or stdin or exit, return a list of residues

    file_name: a path to read the nef from Path('-') read from stdin
    :return: a list of parsed residues and chains, in the order they were read
    """
    try:
        result = get_sequence(file_name)
    except Exception as e:
        msg = "couldn't read sequence from nef input stream, either no stream or no nef molecular system frame"
        exit_error(msg, e)

    if len(result) == 0:
        msg = "no sequence read from nef molecular system in nef input stream"
        exit_error(msg)

    return result


def get_sequence(file_name: Path = Path("-")) -> List[SequenceResidue]:
    """
    read a sequence from a nef molecular system on stdin and return a list of residues

    file_name: a path to read the nef from Path('-') read from stdin
    can raise Exceptions

    :return:a list of parsed residues and chains, in the order they were read
    """

    text = read_from_file_or_exit(file_name, "sequence from NEF molecular system")

    display_file_name = get_display_file_name(file_name)

    try:
        entry = Entry.from_string(text)
    except Exception as e:
        msg = f"couldn't parse text in {display_file_name} as NEF data"
        exit_error(msg, e)

    return sequence_from_entry(entry)


def sequence_from_entry_or_exit(entry: Entry) -> List[SequenceResidue]:
    sequence = sequence_from_entry(entry)

    if not sequence:
        msg = "did you read a sequence? a sequence is required and was not read"
        exit_error(msg)

    return sequence


def sequence_from_entry(entry) -> List[SequenceResidue]:

    result = []
    if entry is not None:
        frames = entry.get_saveframes_by_category("nef_molecular_system")

        if frames:
            result = sequence_from_frame(frames[0])
    return result


def sequence_from_frame(
    frame: Saveframe, chain_codes_to_select: Union[str, List[str]] = ANY_CHAIN
) -> List[SequenceResidue]:

    """
    read sequences from a nef molecular system save frame
    can raise Exceptions

    :param frame: the save frame to read residues from, must have category nef molecular system
    :param chain_code_to_select: the chain codes to select,t this can be either a string or list of strings,
                                 any chain is selected using the instance of the constant string ANY_CHAIN [default]
    :return: a list of parsed residues and chains, in the order they were read
    """

    if chain_codes_to_select is not ANY_CHAIN and isinstance(
        chain_codes_to_select, str
    ):
        chain_codes_to_select = set(
            [
                chain_codes_to_select,
            ]
        )
    elif chain_codes_to_select is not ANY_CHAIN:
        chain_codes_to_select = set(chain_codes_to_select)

    residues = OrderedSet()

    if frame.category != "nef_molecular_system":
        raise Exception(
            f"sequences can only be read from nef molecular system frames, the category of the provided frame "
            f"was {frame.category}"
        )

    loop = frame.loops[0]

    chain_code_index = loop.tag_index("chain_code")
    sequence_code_index = loop.tag_index("sequence_code")
    residue_name_index = loop.tag_index("residue_name")
    linking_index = loop.tag_index("linking")

    for line in loop:
        chain_code = line[chain_code_index]
        if chain_codes_to_select is ANY_CHAIN or chain_code in chain_codes_to_select:
            sequence_code = line[sequence_code_index]
            residue_name = line[residue_name_index]
            linking = (
                Linking[line[linking_index].upper()]
                if line[linking_index] != NEF_UNKNOWN
                else None
            )
            residue = SequenceResidue(
                chain_code=chain_code,
                sequence_code=sequence_code,
                residue_name=residue_name,
                linking=linking,
            )
            residues.append(residue)

    return list(residues)


def sequence_3let_to_res(
    sequence_3_let: List[str], chain_code: str, start: int = 1
) -> List[SequenceResidue]:
    """
    convert a list of 3 letter residue names to a list of sequence residues
    :param sequence_3_let: 3 letter names
    :param chain_code: chain code to use
    :param start:  start of the chain [default is 1]
    :return: a list of sequeence residues
    """
    result = set()
    for res_number, residue_name in enumerate(sequence_3_let, start=start):
        result.add(SequenceResidue(chain_code, res_number, residue_name))

    return list(result)


def get_chain_starts(residues: List[SequenceResidue]) -> Dict[str, int]:
    """
    from a list of residues get the lowest residue number for each chain ignoring any sequence codes that
    can't be converted to an integer

    :param residues:  a list of residues from one ore more chains
    :return: a dictionary of chain starts by chain_code
    """

    chain_residue_numbers = {}

    for residue in residues:

        if is_int(residue_number := residue.sequence_code):
            chain_residue_numbers.setdefault(residue.chain_code, []).append(
                residue_number
            )

    return {
        chain_code: min(residue_numbers)
        for chain_code, residue_numbers in chain_residue_numbers.items()
    }


def sequence_to_chains(residues: List[SequenceResidue]) -> List[str]:
    """
    from a list of residues get chain_codes

    :param residues:  a list of residues from one ore more chains
    :return: a list of chain_codes
    """

    chain_codes = set()

    for residue in residues:

        chain_codes.add(residue.chain_code)

    return sorted(tuple(chain_codes))


def offset_chain_residues(
    residues: List[SequenceResidue], chain_residue_offsets: Dict[str, int]
) -> List[SequenceResidue]:
    """
    take a list of residues and offset residue sequence_codes by any offsets in chains if the
    sequence_code can be converted to an integer and the chain_code matches

    :param residues:  a list of residues from one ore more chains
    :return: a list of residues
    """

    result = []
    for residue in residues:
        if residue.chain_code in chain_residue_offsets and is_int(
            residue.sequence_code
        ):
            offset = chain_residue_offsets[residue.chain_code]
            result.append(
                replace(residue, sequence_code=residue.sequence_code + offset)
            )
        else:
            result.append(residue)

    return result


def make_chunked_sequence_1let(
    sequence_1_let: List[str], sub_chunk: int = 10, line_length: int = 100
) -> List[str]:
    """
    convert a list of strings into a chunked list of strings with (by de]fault) every 100 sting being
    separated by a new line [line_length] and sequences of strings being separated by a space
    :param sequence_1_let: a list of strings typically 1 single letter
    :param sub_chunk: how often to separate strings by a space
    :param line_length: how many strings need to be seen before a new line is added
    :return: a list of formatted strings one per line
    """

    rows = chunks(sequence_1_let, 100)

    row_strings = []
    for row in rows:
        row_chunks = list(chunks(row, 10))
        row_string = ["".join(chunk) for chunk in row_chunks]
        row_strings.append(" ".join(row_string))

    return row_strings


def residues_to_residue_name_lookup(
    sequence: List[SequenceResidue],
) -> Dict[Tuple[str, str], str]:
    return {
        (residue.chain_code, str(residue.sequence_code)): residue.residue_name
        for residue in sequence
    }


def replace_chain_in_atom_labels(
    atom_labels: List[AtomLabel], chain_code: str
) -> List[AtomLabel]:
    """
    replace the chain in a list of AtomLabels, not this operation is not inplace new AtomLabels are returned

    :param atom_labels: a list of AtomLabels
    :param chain_code: the new chain_code
    :return: the updated AtomLabels
    """
    result = []

    for selection in atom_labels:
        new_residue = replace(selection.residue, chain_code=chain_code)
        result.append(replace(selection, residue=new_residue))

    return result


def get_residue_name_from_lookup(
    chain_code: str,
    sequence_code: Union[str, int],
    lookup_table: Dict[Union[str, int], str],
) -> str:
    """
    get a residue name from a chain_code and sequence_code, using the provided lookup table. The sequence code is
    treated being as either  a string or an int, strs are checked first
    :param chain_code: the chain code
    :param sequence_code: the residue code treated as both an int and a string
    :param lookup_table: the table of residues names indexed by chain_code_sequence_code
    :return:
    """

    seq_key = chain_code, str(sequence_code)

    residue_name = lookup_table.get(seq_key, None)

    if not residue_name:
        if is_int(sequence_code):
            seq_key = chain_code, int(sequence_code)

    residue_name = lookup_table.get(seq_key, UNUSED)

    if residue_name is not None:
        residue_name = residue_name.upper()

    return residue_name


def sequence_to_residue_type_lookup(
    sequence: List[SequenceResidue],
) -> Dict[Tuple[str, int], str]:

    """
    build a lookup table from sequence_code, chain_code to residue_type from a list of sequence residues
    :param sequence: a list of sequence residues
    :return: the lookup table
    """
    result: Dict[Tuple[str, int], str] = {}
    for residue in sequence:
        result[residue.chain_code, residue.sequence_code] = residue.residue_name
    return result



import sys
from datetime import date
from textwrap import dedent, indent

import typer

from nef_pipelines import nef_app
from nef_pipelines.lib.util import get_version

if nef_app:
    # noinspection PyUnusedLocal
    @nef_app.app.command()
    def about(
        version: bool = typer.Option(
            False, "--version", help="only display the current version"
        )
    ):
        "- info about NEF-Pipelines"

        print(cmd(version), file=sys.stderr)

    def cmd(version):
        version_data = get_version()

        if version:
            result = str(version_data)
        else:
            NEF_PIPLINES_URL = "https://github.com/varioustoxins/NEF-Pipelines"
            msg = f"""\

                This is NEF-Pipelines version {version_data}

                For information about this program and the code please see: {NEF_PIPLINES_URL}
                For bug reporting please goto: {NEF_PIPLINES_URL}/issues

                If you like this project please star it on the projects github page (listed above)!
                If you want to cite the project please cite it as

                G. S. Thompson {NEF_PIPLINES_URL} ({date.today().year})

                regards Gary (aka varioustoxins)
            """

            msg = dedent(msg)

            max_width = max([len(line) for line in msg.split("\n")])

            header = "=" * (max_width + 4)

            result = "\n".join([header, indent(msg, "  "), header])

        return result



from textwrap import dedent

import typer
from typer.testing import CliRunner

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.pales.exporters.rdcs import rdcs

runner = CliRunner()
app = typer.Typer()
app.command()(rdcs)

NEF_STREAM = open(path_in_test_data(__file__, "pales_test_1.nef")).read()
NEF_STREAM_DISORDERED = open(path_in_test_data(__file__, "pales_test_2.nef")).read()
NEF_STREAM_SEGIDS = open(path_in_test_data(__file__, "pales_test_segids.nef")).read()


# noinspection PyUnusedLocal
def test_rdcs(clear_cache):

    result = run_and_report(app, [], input=NEF_STREAM)

    if result.exit_code != 0:
        print("INFO: stdout from failed read:\n", result.stdout)

    assert result.exit_code == 0

    EXPECTED = """\

        VARS    RESID_I  RESNAME_I  ATOMNAME_I  RESID_J  RESNAME_J  ATOMNAME_J  D      DD     W
        FORMAT  %5d      %6s        %6s         %5d      %6s        %6s         %9.3f  %9.3f  %.2f
                3        GLY        HN          22       GLY        N           3.1    0.4    1.0
                2        TRP        HN          21       TRP        N          -5.2    0.33   1.0

    """
    EXPECTED = dedent(EXPECTED)

    assert_lines_match(EXPECTED, result.stdout)


# noinspection PyUnusedLocal
def test_rdcs_disordered(clear_cache):

    result = run_and_report(app, [], input=NEF_STREAM_DISORDERED)

    assert result.exit_code == 0

    EXPECTED = """\

        VARS    RESID_I  RESNAME_I  ATOMNAME_I  RESID_J  RESNAME_J  ATOMNAME_J  D      DD     W
        FORMAT  %5d      %6s        %6s         %5d      %6s        %6s         %9.3f  %9.3f  %.2f
                3        GLY        HN          22       GLY        N           3.1    0.4    1.0
                2        TRP        HN          21       TRP        N          -5.2    0.33   1.0

    """

    assert_lines_match(EXPECTED, result.stdout)


def test_rdcs_segid(clear_cache):

    result = run_and_report(app, ["--segids"], input=NEF_STREAM_SEGIDS)

    if result.exit_code != 0:
        print("INFO: stdout from failed read:\n", result.stdout)

    assert result.exit_code == 0

    EXPECTED = """\

        VARS    SEGNAME_I RESID_I  RESNAME_I  ATOMNAME_I  SEGNAME_J RESID_J  RESNAME_J  ATOMNAME_J  D      DD     W
        FORMAT  %4s       %5d      %6s        %6s         %4s       %5d      %6s        %6s         %9.3f  %9.3f  %.2f
                AAAA      3        GLY        HN          AAAA      22       GLY        N           3.1    0.4    1.0
                AAAA      2        TRP        HN          AAAA      21       TRP        N          -5.2    0.33   1.0

    """
    EXPECTED = dedent(EXPECTED)

    assert_lines_match(EXPECTED, result.stdout)



import typer

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.tools.frames.list import list

app = typer.Typer()
app.command()(list)


# noinspection PyUnusedLocal
def test_frame_basic():

    path = path_in_test_data(__file__, "frames.nef")
    result = run_and_report(app, ["--in", path])

    EXPECTED = """\
        entry test

        nef_nmr_meta_data  nef_molecular_system
    """

    assert_lines_match(EXPECTED, result.stdout)


# noinspection PyUnusedLocal
def test_frame_file_first():

    path = path_in_test_data(__file__, "frames.nef")

    result = run_and_report(app, [path])

    EXPECTED = """\
        entry test

        nef_nmr_meta_data  nef_molecular_system
    """

    assert_lines_match(EXPECTED, result.stdout)


# noinspection PyUnusedLocal
def test_frame_file_first_and_select():

    path = path_in_test_data(__file__, "frames.nef")
    result = run_and_report(app, [path, "meta_data"])

    EXPECTED = """\
        entry test

        nef_nmr_meta_data
    """

    assert_lines_match(EXPECTED, result.stdout)



import typer

from nef_pipelines import nef_app

app = typer.Typer()
import_app = typer.Typer()
export_app = typer.Typer()


if nef_app.app:

    nef_app.app.add_typer(
        app,
        name="sparky",
        help="-  read sparky files [shifts]",
    )

    app.add_typer(import_app, name="import", help="-  import sparky [shifts]")
    app.add_typer(export_app, name="export", help="-  export sparky [peaks]")

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.transcoders.sparky.exporters.peaks  # noqa: F401
    import nef_pipelines.transcoders.sparky.importers.peaks  # noqa: F401
    import nef_pipelines.transcoders.sparky.importers.shifts  # noqa: F401



import pytest

from nef_pipelines.lib.structures import SequenceResidue
from nef_pipelines.lib.test_lib import path_in_test_data
from nef_pipelines.transcoders.xplor.psf_lib import PSFParseException, parse_xplor_PSF

A3_AB_PSF = open(path_in_test_data(__file__, "3a_ab.psf")).read()


def test_parse_3a_ab():
    result = parse_xplor_PSF(A3_AB_PSF)

    assert len(result) == 6

    EXPECTED = tuple(
        [
            SequenceResidue(chain_code="AAAA", sequence_code=1, residue_name="ALA"),
            SequenceResidue(chain_code="AAAA", sequence_code=2, residue_name="ALA"),
            SequenceResidue(chain_code="AAAA", sequence_code=3, residue_name="ALA"),
            SequenceResidue(chain_code="BBBB", sequence_code=11, residue_name="ALA"),
            SequenceResidue(chain_code="BBBB", sequence_code=12, residue_name="ALA"),
            SequenceResidue(chain_code="BBBB", sequence_code=13, residue_name="ALA"),
        ]
    )

    assert EXPECTED == result


EMPTY_PSF = open(path_in_test_data(__file__, "empty.psf")).read()


def test_parse_empty():
    with pytest.raises(PSFParseException) as excinfo:
        parse_xplor_PSF(EMPTY_PSF, "test-empty-file.psf")

    exception_message = excinfo.value.args[0]

    print(exception_message)

    assert "test-empty-file.psf" in exception_message
    assert "the first line of a PSF file should be 'PSF'" in exception_message


BAD_NATOM_FIELDS_PSF = open(path_in_test_data(__file__, "bad_natom.psf")).read()


def test_parse_bad_natom():
    with pytest.raises(PSFParseException) as excinfo:
        parse_xplor_PSF(BAD_NATOM_FIELDS_PSF, "bad_natom.psf")

    exception_message = excinfo.value.args[0]

    assert "bad_natom.psf" in exception_message
    assert "can't convert natom to an int" in exception_message


BAD_HEADER_PSF = open(path_in_test_data(__file__, "bad_header.psf")).read()


def test_parse_bad_header():
    with pytest.raises(PSFParseException) as excinfo:
        parse_xplor_PSF(BAD_HEADER_PSF, "bad_header.psf")

    exception_message = excinfo.value.args[0]

    assert "bad_header.psf" in exception_message
    assert (
        "the first line of a PSF file should be 'PSF' i got 'WIBBLE' "
        in exception_message
    )


BAD_FIELD_COUNT = open(path_in_test_data(__file__, "bad_field_count.psf")).read()


def test_parse_bad_field_count():
    with pytest.raises(PSFParseException) as excinfo:
        parse_xplor_PSF(BAD_FIELD_COUNT, "bad_field_count.psf")

    exception_message = excinfo.value.args[0]

    assert "8 fields" in exception_message
    assert "atom number 1" in exception_message
    assert "expected 9" in exception_message


BAD_RESID_FIELD = open(path_in_test_data(__file__, "bad_residue_number.psf")).read()


def test_parse_bad_resid():
    with pytest.raises(PSFParseException) as excinfo:
        parse_xplor_PSF(BAD_RESID_FIELD, "bad_residue_number.psf")

    exception_message = excinfo.value.args[0]

    assert "bad_residue_number" in exception_message
    assert "couldn't convert" in exception_message
    assert "ZZZ" in exception_message


NO_RESIDUES_FOUND = open(path_in_test_data(__file__, "no_residues_found.psf")).read()


def test_parse_no_residues_found():
    with pytest.raises(PSFParseException) as excinfo:
        parse_xplor_PSF(NO_RESIDUES_FOUND, "no_residues_found.psf")

    exception_message = excinfo.value.args[0]

    assert "no residues found" in exception_message



import string
import sys
from pathlib import Path
from textwrap import dedent
from typing import Dict, List

import typer

from nef_pipelines.lib.isotope_lib import ATOM_TO_ISOTOPE
from nef_pipelines.lib.nef_lib import (
    UNUSED,
    add_frames_to_entry,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.lib.peak_lib import peaks_to_frame
from nef_pipelines.lib.sequence_lib import TRANSLATIONS_1_3
from nef_pipelines.lib.structures import (
    AtomLabel,
    LineInfo,
    NewPeak,
    Residue,
    ShiftData,
)
from nef_pipelines.lib.util import (
    STDIN,
    exit_error,
    is_float,
    parse_comma_separated_options,
)
from nef_pipelines.transcoders.sparky import import_app


# TODO: this needs to be moved to a library
class SparkyPeakListException(Exception):
    pass


class IncompatibleDimensionTypesException(SparkyPeakListException):
    pass


app = typer.Typer()

DEFAULT_NUCLEI_HELP = (
    "nuclei to use for each dimension if not defined they are guessed from the assigments"
    "or an error is reported"
)


# noinspection PyUnusedLocal
@import_app.command(no_args_is_help=True)
def peaks(
    # chain_codes: List[str] = typer.Option(
    #     None,
    #     "--chains",
    #     help="chain codes as a list of names separated by commas, repeated calls will add further chains [default A]",
    #     metavar="<CHAIN-CODES>",
    # ),
    frame_name: str = typer.Option(
        "sparky", "-f", "--frame-name", help="a name for the frame"
    ),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--input",
        metavar="|PIPE|",
        help="input to read NEF data from [- is stdin]",
    ),
    file_names: List[Path] = typer.Argument(
        ..., help="input files of type peaks.txt", metavar="<SPARKY-peaks>.txt"
    ),
    nuclei: List[str] = typer.Option([], help=DEFAULT_NUCLEI_HELP),
    default_chain_code: str = typer.Option(
        "A",
        "-c",
        "--chain-code",
        help="default chain code to use if none is provided in the file",
    ),
    spectrometer_frequency: float = typer.Option(
        600.123456789, help="spectrometer frequency in MHz"
    ),
):
    """convert sparky peaks file <SPARKY-peaks>.txt to NEF [alpha]"""

    print(
        "*** WARNING *** this command [sparky peaks] is only lightly tested use at your own risk!",
        file=sys.stderr,
    )

    # chain_codes = parse_comma_separated_options(chain_codes)
    #
    # if not chain_codes:
    #     chain_codes = ["A"]

    # _exit_if_chain_codes_and_file_name_dont_match(chain_codes, file_names)

    entry = read_or_create_entry_exit_error_on_bad_file(input)

    # sequence = sequence_from_entry_or_exit(entry)

    nuclei = parse_comma_separated_options(nuclei)

    pipe(
        entry,
        frame_name,
        file_names,
        input_dimensions=nuclei,
        spectrometer_frequency=spectrometer_frequency,
    )


def parse_header_to_columns(header_line: str, file_name) -> Dict[str, int]:
    headings_to_columns = {}
    headings = header_line.split()

    if "Data" in headings:
        headings.remove("Data")

    while ("hz") in headings:
        headings.remove("hz")

    for i, heading in enumerate(headings):
        if heading in "Assignment Height Volume".split():
            headings_to_columns[heading] = i
            continue

        if heading.startswith("w") or heading.startswith("lw"):
            headings_to_columns[heading] = i
            continue

        msg = f"""
            WARNING: unexpected heading {heading} in the file {file_name}...
                     this heading will be ignored, please send the heading and first
                     few lines of this file to the developers of NEF-Pipelines
        """

        print(msg, file=sys.stderr)

    assignment_column = headings_to_columns["Assignment"]
    if assignment_column != 0:
        msg = f"""
            The file {file_name} doesn't look like a sparky file the Assignment column should be the first column
            i got {assignment_column}
            header was
            {header_line}
        """
        exit(msg)

    return headings_to_columns


def _exit_error_if_shift_not_float(shifts, line_info):
    for i, shift in enumerate(shifts, start=1):
        if not is_float(shift):
            msg = f"""
                file {line_info.file_name} does not look like a spark file
                for shift w{i} at line {line_info.line_no} the value {shift} couldn't be converted to a float
                the full line was
                {line_info.line}
            """
            exit_error(msg)


def _parse_peaks(lines, file_name):

    peaks = []

    in_data = False

    dimension_count = None

    for line_number, line in enumerate(lines, start=1):

        line_info = LineInfo(file_name, line_number, line)
        line = line.strip()
        if len(line) == 0:
            continue

        if line.startswith("#"):
            continue

        if line.strip()[: len("Assignment")] == "Assignment":
            if in_data:
                _exit_error_header_in_data(line_info)

            column_headers_to_indices = parse_header_to_columns(line, file_name)

            dimension_count = _count_dimensions(column_headers_to_indices)

            _exit_error_no_dimensions(dimension_count, line_info)

            in_data = True

            continue
        else:
            if not in_data:
                _exit_error_data_but_no_header(line_info)

            fields = line.split()
            values = {}
            for column, index in column_headers_to_indices.items():
                if index < len(fields):
                    values[column] = fields[index]
                else:
                    _exit_error_not_enough_columns_in_data_row(
                        column_headers_to_indices, column, line_info
                    )

            assignmnents_column = column_headers_to_indices["Assignment"]
            raw_assignment = fields[assignmnents_column]

            assignments = _process_assignments(raw_assignment)

            shifts = [
                fields[column_headers_to_indices[f"w{index}"]]
                for index in range(1, dimension_count + 1)
            ]

            shifts = [float(shift) for shift in shifts if is_float(shift)]

            _exit_error_if_shift_not_float(shifts, line_info)

            shift_data = [
                ShiftData(atom=atom, value=value)
                for atom, value in zip(assignments, shifts)
            ]
            peak = NewPeak(
                shifts=shift_data,
            )

            peaks.append(peak)

    return peaks


def _exit_error_data_but_no_header(line_info):
    msg = f"""
                    the file {line_info.file_name} doesn't look like a sparky file at line {line_info.line_no}
                    there appears to be a data line but no header was detected a head should look like
                    Assignment w1 w2 ... etc
                    the current line data is
                    {line_info.line}
                """
    exit_error(msg)


def _exit_error_not_enough_columns_in_data_row(
    column_headers_to_indices, column, line_info
):
    msg = f"""
                        In sparky peaks file {line_info.file_name} at line {line_info.line_number} for column {column}
                        there were was not enough data [expected {len(column_headers_to_indices)} columns]
                        the line was:
                        {line_info.line}
                    """
    exit_error(msg)


def _exit_error_no_dimensions(dimension_count, line_info):
    if dimension_count < 1:
        msg = f"""
                    sparky peak file {line_info.file_name} doesn't appear to have enough columns [minimum 1]
                    at line {line_info.line_no}
                    the header was:
                    {line_info.line}
                """
        sys.exit(msg)


def _count_dimensions(column_headers_to_indices):
    dimension_count = 0
    for name in [f"w{i}" for i in range(1, 20)]:
        if name in column_headers_to_indices:
            dimension_count += 1
        else:
            break
    return dimension_count


def _exit_error_header_in_data(line_info):
    msg = f"""
        bad sparky peak file {line_info.file_name} at line no {line_info.line_number} there appears to be a
        second header...
        the line was:
        {line_info.line}
                """
    exit_error(msg)


# TODO this doesn't cope with abbreviated names
def _process_assignments(assignments, chain_code="A"):

    fields = assignments.split("-")

    assignments = []
    for field in fields:

        if len(field) == 1 and field == "?":
            assignment = AtomLabel(
                atom_name=UNUSED, residue=Residue("@-", UNUSED, UNUSED)
            )
            assignments.append(assignment)
        else:

            without_first_letters = field.lstrip(string.ascii_letters)
            first_letters = field[: -len(without_first_letters)]
            if len(first_letters) == 0 and without_first_letters[0] == "?":
                first_letters = "?"
                without_first_letters = without_first_letters[1:]

            without_numbers = without_first_letters.lstrip(string.digits)
            numbers = without_first_letters[: -len(without_numbers)]

            if len(numbers) == 0 and without_numbers[0] == "?":
                numbers = "?"
                without_numbers = without_numbers[1:]

            if len(first_letters) == 0 or first_letters == "?":
                residue_name = UNUSED
            else:
                residue_name = first_letters

            if len(numbers) == 0 or numbers == "?":
                sequence_code = UNUSED
            else:
                sequence_code = numbers

            if len(without_numbers) == 0 or without_numbers == "?":
                atom_name = UNUSED
            else:
                atom_name = without_numbers

            if residue_name in TRANSLATIONS_1_3:
                residue_name = TRANSLATIONS_1_3[residue_name]

            assignment = AtomLabel(
                atom_name=atom_name,
                residue=Residue(chain_code, sequence_code, residue_name),
            )
            assignments.append(assignment)

    return assignments


def pipe(
    entry, frame_name, file_names, input_dimensions, spectrometer_frequency
):  # , sequence, chain_codes,

    sparky_frames = []

    for (
        file_name
    ) in file_names:  # , chain_code in zip(file_names, chain_code_iter(chain_codes)):

        # with cached_file_stream(file_name) as lines:
        #
        #     chain_seqid_to_type = sequence_to_residue_type_lookup(sequence)

        try:
            with open(file_name, "r") as fp:
                lines = fp.readlines()
        except IOError as e:
            msg = f"""
                    while reading sparky peaks file {file_name} there was an error reading the file
                    the error was: {e}
                """
            exit_error(msg, e)

        sparky_peaks = _parse_peaks(
            lines, file_name=file_name  # chain_seqid_to_type, chain_code=chain_code,
        )

        dimensions = _guess_dimensions_if_not_defined_or_throw(
            sparky_peaks, input_dimensions
        )

        dimensions = [{"axis_code": dimension} for dimension in dimensions]

        frame = peaks_to_frame(sparky_peaks, dimensions, spectrometer_frequency)

        sparky_frames.append(frame)
    #
    entry = add_frames_to_entry(entry, sparky_frames)

    print(entry)


# TODO: this needs to be moved to a library
def _guess_dimensions_if_not_defined_or_throw(
    peaks: List[NewPeak], input_dimensions: List[str]
) -> List[str]:
    results_by_dimension = {
        i: dimension for i, dimension in enumerate(input_dimensions)
    }

    guessed_dimensions = {}

    for peak_number, peak in enumerate(peaks, start=1):

        for dimension_index, shift in enumerate(peak.shifts):
            atom_name = shift.atom.atom_name

            if len(atom_name) == 0:
                atom_name = None

            atom_type = None
            if atom_name != UNUSED and atom_name is not None:
                atom_type = atom_name.strip()[0]

            if atom_type is not None and atom_type.lower() not in string.ascii_letters:
                atom_type = None

            guessed_dimension_isotopes = guessed_dimensions.setdefault(
                dimension_index, set()
            )
            if atom_type and atom_type in ATOM_TO_ISOTOPE:
                isotope_code = ATOM_TO_ISOTOPE[atom_type]
                guessed_dimension_isotopes.add(isotope_code)
            else:
                guessed_dimension_isotopes.add(UNUSED)

    for guessed_dimension_index in guessed_dimensions:
        if guessed_dimension_index not in results_by_dimension:

            guessed_dimension = guessed_dimensions[guessed_dimension_index]
            if len(guessed_dimension) > 1 and UNUSED in guessed_dimension:
                guessed_dimension.remove(UNUSED)

            if len(guessed_dimensions[guessed_dimension_index]) > 1:

                _raise_multiple_guessed_isotopes(
                    guessed_dimensions, guessed_dimension_index
                )

    for dimension_index in guessed_dimensions:
        if dimension_index not in results_by_dimension:
            results_by_dimension[dimension_index] = list(
                guessed_dimensions[dimension_index]
            )[0]

    max_dimension = max(results_by_dimension.keys())

    result = []
    for i in range(max_dimension + 1):
        result.append(results_by_dimension[i])

    return result


def _raise_multiple_guessed_isotopes(guessed_dimensions, guessed_dimension_index):
    dim_info = []
    for dimension_index in sorted(guessed_dimensions).keys():
        dim_data = f"{dimension_index}: {' '.join(guessed_dimensions.values())}"
        dim_info.append(dim_data)
    msg = f"""\
                    no isotope provided for dimension {guessed_dimension_index + 1} and multiple possibilities
                    found in atom names, please use explicit isotope codes using dimension-nuclei

                    deduced codes for dimensions:


                """
    msg = dedent(msg)
    msg += dim_info
    raise IncompatibleDimensionTypesException(msg)
